"""Locales for naming conventions."""

import gettext
from pathlib import Path

from attrs import define, field

LOCALES_DIR = Path(__file__).parent


@define(frozen=True)
class Locales:
    translations: gettext.GNUTranslations = field()

    @classmethod
    def from_domain(cls, domain: str, locales_dir: Path = LOCALES_DIR):
        translations = gettext.translation(
            domain,
            localedir=locales_dir,
            languages=["en"],
        )
        return cls(translations)

    def translate(self, message: str, context: str = "") -> str:
        if not message:
            return message

        if context:
            new_message = self.translations.pgettext(context, message)
            if new_message != message:
                return new_message

        return self.translations.gettext(message)



"""Geocoding data from OpenStreetMap."""

from functools import partial

from attrs import define, evolve, field

from permaculture.http import HTTPSession
from permaculture.location import (
    LocationMultiPolygon,
    LocationNotFound,
    LocationPoint,
)

NOMINATIM_ORIGIN = "https://nominatim.openstreetmap.org"


@define(frozen=True)
class Nominatim:
    """Nominatim uses OpenStreetMap data to find locations on Earth."""

    session = field(factory=partial(HTTPSession, NOMINATIM_ORIGIN))

    def with_cache(self, storage):
        session = self.session.with_cache(storage)
        return evolve(self, session=session)

    def search(self, **params):
        """Search nominatim with the given params."""
        params.setdefault("format", "json")
        return self.session.get("/search", params=params).json()

    def point(self, query):
        """Get the point of a place."""
        data = self.search(q=query)
        if not data:
            raise LocationNotFound(f"No data for {query!r}")

        # TODO: validate what happens when len(data) > 1
        first = data[0]
        return LocationPoint(first["lat"], first["lon"])

    def multi_polygon(self, query):
        """Get a multiple polygons around a place."""
        data = self.search(
            q=query,
            format="geojson",
            polygon_geojson="1",
        )
        if not data:
            raise LocationNotFound(f"No data for {query!r}")

        coordinates = data["features"][0]["geometry"]["coordinates"]
        # Geojson points use [longitude, latitude].
        # https://en.wikipedia.org/wiki/GeoJSON
        return LocationMultiPolygon.from_polygons(
            [
                [(point[1], point[0]) for point in poly]
                for multi in coordinates
                for poly in multi
            ]
        )



"""Unit tests for the Google module."""

from unittest.mock import Mock

from yarl import URL

from permaculture.google import GoogleSpreadsheet

from .stubs import StubRequestsResponse


def test_google_spreadsheet_from_url(unique):
    """Instantiating from URL should parse the doc ID."""
    doc_id = unique("text")
    url = URL(f"https://docs.google.com/spreadsheets/d/{doc_id}/edit")
    gs = GoogleSpreadsheet.from_url(url)
    assert gs.doc_id == doc_id


def test_google_spreadsheet_export(unique):
    """Exporting should GET with gid and format params."""
    doc_id, gid, fmt = unique("text"), unique("text"), unique("text")
    session = Mock(post=Mock(return_value=StubRequestsResponse()))
    GoogleSpreadsheet(session, doc_id).export(gid, fmt)
    session.get.assert_called_once_with(
        f"/spreadsheets/d/{doc_id}/export",
        params={"gid": gid, "format": fmt},
    )



"""Strategy for calculating the priority."""

from functools import cached_property

from attrs import define, evolve, field

from permaculture.ipinfo import IPinfo
from permaculture.nominatim import Nominatim


@define(frozen=True)
class Priority:
    weight = 1.0


@define(frozen=True, slots=False)
class LocationPriority(Priority):
    query = field()
    nominatim = field(factory=Nominatim)
    ipinfo = field(factory=IPinfo)

    def with_cache(self, storage):
        nominatim = self.nominatim.with_cache(storage)
        ipinfo = self.ipinfo.with_cache(storage)
        return evolve(self, nominatim=nominatim, ipinfo=ipinfo)

    @cached_property
    def weight(self):
        multi = self.nominatim.multi_polygon(self.query)
        point = self.ipinfo.point()
        distance = multi.distance(point)
        return 1 / distance if distance > 1 else 1.0



"""Unit tests for the testing unique module."""

from http.cookiejar import Cookie, CookieJar

import pytest
from hamcrest import assert_that, is_

from permaculture.testing.unique import (
    unique_cookie,
    unique_cookies,
)


@pytest.mark.parametrize(
    "name, matches",
    [
        ("cookie", is_(Cookie)),
        ("cookies", is_(CookieJar)),
    ],
)
def test_unique_types(name, matches, unique):
    """Unique fixtures should return expected types."""
    assert_that(unique(name), matches)


@pytest.mark.parametrize(
    "plugin",
    [
        pytest.param(unique_cookie, id="cookie"),
        pytest.param(unique_cookies, id="cookies"),
    ],
)
def test_unique_plugin_twice(plugin, unique):
    """Calling any plugin twice should not return the same value."""
    assert plugin(unique) != plugin(unique)


@pytest.mark.parametrize(
    "count",
    [
        0,
        1,
        2,
    ],
)
def test_unique_cookies_count(count, unique):
    """Passing a count should return that many cookies."""
    assert len(unique_cookies(unique, count)) == count



"""Testing fixtures.

The logger_handler fixture is automatically provided by this package,
it does not need to be imported in conftest.py explicitly.
"""

import logging

import pytest

from permaculture.logger import setup_logger
from permaculture.testing.logger import LoggerHandler


@pytest.fixture(autouse=True)
def logger_handler():
    """Logger handler fixture."""
    handler = LoggerHandler()
    setup_logger(logging.DEBUG, handler)
    try:
        yield handler
    finally:
        setup_logger()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from importlib.metadata import distribution

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "permaculture"
copyright = "2023, Marc Tardif"  # noqa: A001
author = "Marc Tardif"
version = release = distribution(project).version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinxcontrib.log_cabinet",
    "sphinx_rtd_theme",
]

language = "en"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"



"""Unit tests for the action module."""

from argparse import ArgumentParser
from enum import Enum

import pytest
from hamcrest import assert_that, has_properties

from permaculture.action import SingleAction, enum_action

StubEnum = Enum("StubEnum", ["a", "b"])


def test_enum_action():
    """An EnumAction should return an enum instance."""
    parser = ArgumentParser()
    parser.add_argument("--enum", action=enum_action(StubEnum))

    result = parser.parse_args(["--enum", "a"])

    assert_that(result, has_properties(enum=StubEnum.a))


def test_enum_action_default():
    """An EnumAction should return the default when specified."""
    parser = ArgumentParser()
    parser.add_argument(
        "--enum",
        action=enum_action(StubEnum),
        default=StubEnum.a,
    )

    result = parser.parse_args([])

    assert_that(result, has_properties(enum=StubEnum.a))


def test_enum_action_list():
    """An EnumAction should return a list of values."""
    parser = ArgumentParser()
    parser.add_argument("--enum", action=enum_action(StubEnum), nargs="*")

    result = parser.parse_args(["--enum", "a", "b"])

    assert_that(result, has_properties(enum=[StubEnum.a, StubEnum.b]))


def test_single_action(unique):
    """A SingleAction should allow a single action."""
    parser = ArgumentParser()
    parser.add_argument("--action", action=SingleAction)

    action = unique("text")
    result = parser.parse_args(["--action", action])

    assert_that(result, has_properties(action=action))


def test_single_action_value_error():
    """A SingleAction with nargs should raise a ValueError."""
    parser = ArgumentParser()

    with pytest.raises(ValueError):
        parser.add_argument("--action", action=SingleAction, nargs=1)



"""USDA Plants database."""

import logging
from functools import partial

from attrs import define, field
from requests.exceptions import HTTPError

from permaculture.converter import Converter
from permaculture.database import DatabasePlant
from permaculture.http import HTTPSession
from permaculture.locales import Locales
from permaculture.priority import LocationPriority, Priority
from permaculture.unit import fahrenheit, feet, inches

USDA_ORIGIN = "https://plantsservices.sc.egov.usda.gov"

logger = logging.getLogger(__name__)


@define(frozen=True)
class USDAWeb:
    """USDA Plants web interface."""

    session: HTTPSession = field(
        factory=partial(HTTPSession, USDA_ORIGIN, headers={"Accept": "application/json"}),
    )

    def characteristics_search(self, **kwargs):
        """Search characteristics."""
        payload = {
            "ArtistFirstLetters": None,
            "Artists": None,
            "Cities": None,
            "CopyrightStatuses": None,
            "Counties": None,
            "Countries": None,
            "Durations": None,
            "Field": None,
            "FilterOptions": None,
            "Groups": None,
            "GrowthHabits": None,
            "ImageLocations": None,
            "ImageTypes": None,
            "InvasiveLocations": None,
            "Localities": None,
            "Locations": None,
            "MasterId": -1,
            "NoxiousLocations": None,
            "Offset": None,
            "Provinces": None,
            "SortBy": "sortSciName",
            "TaxonSearchCriteria": None,
            "Text": None,
            "Type": "Characteristics",
            "UnfilteredPlantIds": None,
            "WetlandRegions": None,
            **kwargs,
        }
        response = self.session.post(
            "/api/CharacteristicsSearch",
            json=payload,
        )
        return response.json()

    def plant_search(self, **kwargs):
        """Search plants."""
        payload = {
            "ArtistFirstLetters": None,
            "Artists": None,
            "Cities": None,
            "CopyrightStatuses": None,
            "Counties": None,
            "Countries": None,
            "Durations": None,
            "Field": None,
            "FilterOptions": None,
            "Groups": None,
            "GrowthHabits": None,
            "ImageLocations": None,
            "ImageTypes": None,
            "InvasiveLocations": None,
            "Localities": None,
            "Locations": None,
            "MasterId": -1,
            "NoxiousLocations": None,
            "Offset": -1,
            "Provinces": None,
            "SortBy": None,
            "TaxonSearchCriteria": None,
            "Text": None,
            "Type": "Characteristics",
            "UnfilteredPlantIds": None,
            "WetlandRegions": None,
            **kwargs,
        }
        response = self.session.post(
            "/api/PlantSearch",
            json=payload,
        )
        return response.json()

    def plant_profile(self, symbol):
        """Plant profile for a symbol."""
        response = self.session.get(
            "/api/PlantProfile",
            params={"symbol": symbol},
        )
        return response.json()

    def plant_characteristics(self, Id):
        """Plant characteristics for an identifier."""
        response = self.session.get(f"/api/PlantCharacteristics/{Id}")
        return response.json()


@define(frozen=True)
class USDAConverter(Converter):
    locales: Locales = field(
        factory=partial(Locales.from_domain, "usda")
    )

    def convert_item(self, key, value):
        dispatchers = {
            "AcceptedId": self.convert_ignore,
            "Adapted to Coarse Textured Soils": self.convert_bool,
            "Adapted to Fine Textured Soils": self.convert_bool,
            "Adapted to Medium Textured Soils": self.convert_bool,
            "Berry/Nut/Seed Product": self.convert_bool,
            "Christmas Tree Product": self.convert_ignore,
            "Cold Stratification Required": self.convert_bool,
            "CommonName": self.convert_list,
            "Coppice Potential": self.convert_bool,
            "Fall Conspicuous": self.convert_bool,
            "Fire Resistant": self.convert_bool,
            "Flower Conspicuous": self.convert_bool,
            "Flower Color": self.convert_list,
            "Fodder Product": self.convert_bool,
            "Foliage Color": self.convert_list,
            "Frost Free Days, Minimum": self.convert_int,
            "Fruit/Seed Conspicuous": self.convert_bool,
            "Fruit/Seed Persistence": self.convert_bool,
            "Growth Rate": self.convert_list,
            "HasCharacteristics": self.convert_ignore,
            "HasDistributionData": self.convert_ignore,
            "HasDocumentation": self.convert_ignore,
            "HasEthnobotany": self.convert_ignore,
            "HasImages": self.convert_ignore,
            "HasInvasiveStatuses": self.convert_ignore,
            "HasLegalStatuses": self.convert_ignore,
            "HasNoxiousStatuses": self.convert_ignore,
            "HasPollinator": self.convert_ignore,
            "HasRelatedLinks": self.convert_ignore,
            "HasSubordinateTaxa": self.convert_ignore,
            "HasSynonyms": self.convert_ignore,
            "HasWetlandData": self.convert_ignore,
            "HasWildlife": self.convert_ignore,
            "Height at 20 Years, Maximum (feet)": self.convert_ignore,
            "Height, Mature (feet)": partial(self.convert_float, unit=feet),
            "Id": self.convert_ignore,
            "ImageId": self.convert_ignore,
            "Known Allelopath": self.convert_bool,
            "Leaf Retention": self.convert_bool,
            "Low Growing Grass": self.convert_bool,
            "Lumber Product": self.convert_bool,
            "Naval Store Product": self.convert_bool,
            "Nursery Stock Product": self.convert_bool,
            "Palatable Human": self.convert_bool,
            "PlantLocationId": self.convert_ignore,
            "Planting Density per Acre, Maximum": self.convert_int,
            "Planting Density per Acre, Minimum": self.convert_int,
            "Post Product": self.convert_bool,
            "Precipitation, Maximum": self.convert_int,
            "Precipitation, Minimum": self.convert_int,
            "Propagated by Bare Root": self.convert_bool,
            "Propagated by Bulb": self.convert_bool,
            "Propagated by Container": self.convert_bool,
            "Propagated by Corm": self.convert_bool,
            "Propagated by Cuttings": self.convert_bool,
            "Propagated by Seed": self.convert_bool,
            "Propagated by Sod": self.convert_bool,
            "Propagated by Sprigs": self.convert_bool,
            "Propagated by Tubers": self.convert_bool,
            "Pulpwood Product": self.convert_bool,
            "RankId": self.convert_ignore,
            "Resprout Ability": self.convert_bool,
            "Root Depth, Minimum (inches)": partial(
                self.convert_float, unit=inches
            ),
            "ScientificName": self.convert_token,
            "Seed per Pound": self.convert_int,
            "Small Grain": self.convert_bool,
            "Symbol": self.convert_ignore,
            "Synonyms": self.convert_ignore,
            "Temperature, Minimum (°F)": partial(
                self.convert_float, unit=fahrenheit
            ),
            "Veneer Product": self.convert_bool,
            "pH, Maximum": self.convert_float,
            "pH, Minimum": self.convert_float,
        }
        return dispatchers.get(key, self.convert_string)(key, value)


@define(frozen=True)
class USDAModel:
    """USDA Plants model."""

    web: USDAWeb = field(factory=USDAWeb)
    converter: USDAConverter = field(factory=USDAConverter)

    def with_cache(self, storage):
        self.web.session.with_cache(storage)
        return self

    def plant_characteristics(self, plant):
        """Return the characteristics for a single plant."""
        return self.converter.convert(
            {
                **plant,
                **{
                    c["PlantCharacteristicName"]: c["PlantCharacteristicValue"]
                    for c in self.web.plant_characteristics(plant["Id"])
                },
            }
        )

    def plant_search(self, text, field):
        try:
            search = self.web.plant_search(
                Field=field,
                SortBy="sortSciName",
                Text=text,
            )
        except HTTPError:
            return []

        for plant in search["PlantResults"]:
            yield self.plant_characteristics(plant)

    def all_characteristics(self):
        """Return the characteristics for all plants."""
        search = self.web.characteristics_search()
        for plant in search["PlantResults"]:
            yield self.plant_characteristics(plant)


@define(frozen=True)
class USDAIngestor:
    model: USDAModel = field(factory=USDAModel)
    priority: Priority = field(factory=Priority)

    @classmethod
    def from_config(cls, config):
        """Instantiate USDAIngestor from config."""
        model = USDAModel().with_cache(config.storage)
        priority = LocationPriority("United States").with_cache(config.storage)
        return cls(model, priority)

    def fetch_all(self):
        count = 0
        for c in self.model.all_characteristics():
            count += 1
            if count % 100 == 0:
                logger.info("USDA: ingested %d plants", count)
            yield DatabasePlant(c, self.priority.weight)
        logger.info("USDA: ingested %d plants total", count)



"""Ingestor protocol for fetching plant data from remote sources."""

import re
from collections.abc import Iterator
from typing import Protocol, runtime_checkable

from permaculture.database import DatabasePlant
from permaculture.registry import registry_load


@runtime_checkable
class Ingestor(Protocol):
    """Protocol for fetching plant data from a remote source."""

    @classmethod
    def from_config(cls, config): ...

    def fetch_all(self) -> Iterator[DatabasePlant]:
        """Yield all plants from this source."""
        ...


class Ingestors(dict):
    @classmethod
    def load(cls, config=None, registry=None):
        """Load ingestors from registry."""
        if registry is None or "ingestors" not in registry:
            registry = registry_load("ingestors", registry)

        include = re.compile("|".join(config.ingestors), re.I)
        ingestors = {
            k: v.from_config(config)
            for k, v in registry.get("ingestors", {}).items()
            if include.match(k)
        }

        return cls(ingestors)



"""Unit tests for the Design Ecologique module."""

from unittest.mock import Mock

import pytest

from permaculture.database import DatabasePlant
from permaculture.de import (
    DEConverter,
    DEIngestor,
    DEModel,
    DEWeb,
)

from .stubs import StubRequestsResponse


def test_de_web_perenial_plants_list(unique):
    """Perenial plants should GET and return the spreadsheet."""
    doc_id = unique("text")
    text = f"<a href='https://docs.google.com/spreadsheets/d/{doc_id}/edit'>"
    session = Mock(get=Mock(return_value=StubRequestsResponse(text=text)))
    plants = DEWeb(session).perenial_plants_list()
    session.get.assert_called_once_with("/liste-de-plantes-vivaces/")
    assert plants.doc_id == doc_id


def test_de_web_perenial_plants_list_error():
    """Perenial plants should raise when spreadhseet is not found."""
    session = Mock(get=Mock(return_value=StubRequestsResponse()))
    with pytest.raises(KeyError):
        DEWeb(session).perenial_plants_list()


@pytest.mark.parametrize(
    "value, expected",
    [
        (
            "value",
            "value",
        ),
        (
            "*",
            None,
        ),
    ],
)
def test_de_converter_translate(value, expected):
    """Translating * should return None."""
    result = DEConverter().translate(value)
    assert result == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        ("X", True),
        ("*", False),
        ("", None),
    ],
)
def test_de_converter_convert_bool(value, expected):
    """Converting a bool should return the True, False, or None."""
    result = DEConverter().convert_bool("", value)
    assert result == [("", expected)]


@pytest.mark.parametrize(
    "item, expected",
    [
        pytest.param(
            ("key", "1,0 - 2,0"),
            [
                ("key/min", 1.0),
                ("key/max", 2.0),
            ],
            id="comma",
        ),
    ],
)
def test_de_converter_convert_range(item, expected):
    """Converting a range should support single and double values."""
    result = DEConverter().convert_range(*item)
    assert result == expected


@pytest.mark.parametrize(
    "item, expected",
    [
        pytest.param(
            ("Accumulateur de Nutriments", "*"),
            [],
            id="nutriments",
        ),
        pytest.param(
            ("Comestible", "Fl,Fr,Fe,N,G,R,S,JP,T,B"),
            [
                ("edible uses/flower", True),
                ("edible uses/fruit", True),
                ("edible uses/leaf", True),
                ("edible uses/nut", True),
                ("edible uses/seed", True),
                ("edible uses/root", True),
                ("edible uses/sap", True),
                ("edible uses/young shoot", True),
                ("edible uses/stem", True),
                ("edible uses/bulb", True),
            ],
            id="edible uses",
        ),
        pytest.param(
            ("Couleur de feuillage", "V Po Pa P F T J"),
            [
                ("foliage color/green", True),
                ("foliage color/purple", True),
                ("foliage color/variegated", True),
                ("foliage color/pale", True),
                ("foliage color/dark", True),
                ("foliage color/spotted", True),
                ("foliage color/yellow", True),
            ],
            id="foliage color",
        ),
        pytest.param(
            ("Couleur de floraison", "Rg Rs B J O P V Br Bl"),
            [
                ("flower color/red", True),
                ("flower color/pink", True),
                ("flower color/white", True),
                ("flower color/yellow", True),
                ("flower color/orange", True),
                ("flower color/purple", True),
                ("flower color/green", True),
                ("flower color/brown", True),
                ("flower color/blue", True),
            ],
            id="flower color",
        ),
        pytest.param(
            ("Couvre-sol", "*"),
            [],
            id="ground cover",
        ),
        pytest.param(
            ("Cultivars intéressants", "*"),
            [],
            id="cultivars",
        ),
        pytest.param(
            ("Eau", "▁ ▅ █"),
            [
                ("moisture/dry", True),
                ("moisture/moderate", True),
                ("moisture/wet", True),
            ],
            id="moisture",
        ),
        pytest.param(
            ("Haie", "*"),
            [],
            id="hedge",
        ),
        pytest.param(
            ("Hauteur(m)", "0,6 \u2013 1,2"),
            [
                ("height/min", 0.6),
                ("height/max", 1.2),
            ],
            id="height with endash",
        ),
        pytest.param(
            ("Inconvénient", "E D A P Épi V B G Pe"),
            [
                ("inconvenience/expansive", True),
                ("inconvenience/dispersive", True),
                ("inconvenience/allergen", True),
                ("inconvenience/poison", True),
                ("inconvenience/thorny", True),
                ("inconvenience/vigorous", True),
                ("inconvenience/burns", True),
                ("inconvenience/invasive", True),
                ("inconvenience/persistent", True),
            ],
            id="inconvenience",
        ),
        pytest.param(
            ("Intérêt automnale hivernal", "A H"),
            [],
            id="interest",
        ),
        pytest.param(
            ("Largeur(m)", "1,0"),
            [
                ("spread/min", 1.0),
                ("spread/max", 1.0),
            ],
            id="spread with only minimum",
        ),
        pytest.param(
            ("Lien Information", ""),
            [],
            id="information",
        ),
        pytest.param(
            ("Lumière", "○ ◐ ●"),
            [
                ("sun/full", True),
                ("sun/partial", True),
                ("sun/shade", True),
            ],
            id="sun",
        ),
        pytest.param(
            ("Multiplication", "B M D S G St P A É T"),
            [
                ("multiplication/cuttings", True),
                ("multiplication/layering", True),
                ("multiplication/division", True),
                ("multiplication/seeds", True),
                ("multiplication/graft", True),
                ("multiplication/stolon", True),
                ("multiplication/spring", True),
                ("multiplication/autumn", True),
                ("multiplication/summer", True),
                ("multiplication/tuber", True),
            ],
            id="Multiplication",
        ),
        pytest.param(
            ("Période de floraison", "P É A"),
            [
                ("bloom period/min", "spring"),
                ("bloom period/max", "autumn"),
            ],
            id="bloom period",
        ),
        pytest.param(
            ("Période de taille", "AD,AF,P,É,A,T,N"),
            [
                ("pruning period/before budburst", True),
                ("pruning period/after flowering", True),
                ("pruning period/spring", True),
                ("pruning period/summer", True),
                ("pruning period/autumn", True),
                ("pruning period/at all times", True),
                ("pruning period/never prune", True),
            ],
            id="pruning period",
        ),
        pytest.param(
            ("Pollinisateurs", "S G V"),
            [
                ("pollinators/specialists", True),
                ("pollinators/generalists", True),
                ("pollinators/wind", True),
            ],
            id="pollinators",
        ),
        pytest.param(
            ("Racine", "B C D F L P R S T"),
            [
                ("root type/bulb", True),
                ("root type/fleshy", True),
                ("root type/suckering", True),
                ("root type/fasciculated", True),
                ("root type/lateral", True),
                ("root type/taproot", True),
                ("root type/rhizome", True),
                ("root type/superficial", True),
                ("root type/tuber", True),
            ],
            id="root type",
        ),
        pytest.param(
            ("Rythme de croissance", "R M L"),
            [
                ("growth rate/fast", True),
                ("growth rate/medium", True),
                ("growth rate/slow", True),
            ],
            id="growth rate",
        ),
        pytest.param(
            ("Texture du sol", "░ ▒ ▓"),
            [
                ("soil/light", True),
                ("soil/medium", True),
                ("soil/heavy", True),
            ],
            id="soil",
        ),
        pytest.param(
            ("Utilisation écologique", "BR,P,Z"),
            [
                ("ecological use/riparian strip", True),
                ("ecological use/slopes", True),
                ("ecological use/flood zone", True),
            ],
            id="ecological use",
        ),
        pytest.param(
            ("Vie sauvage", "N A"),
            [
                ("wildlife/food", True),
                ("wildlife/shelter", True),
            ],
            id="wildlife",
        ),
        pytest.param(
            ("pH (Min-Max)", "6 - 7"),
            [
                ("ph/min", 6.0),
                ("ph/max", 7.0),
            ],
            id="ph with dash",
        ),
    ],
)
def test_de_converter_convert_item(item, expected):
    """Converting an item should consider types."""
    result = DEConverter().convert_item(*item)
    assert result == expected


def test_de_model_get_perenial_plants():
    """All perenial plants should return a dictionary of characteristics."""
    data = "TAXONOMIE\nGenre,Espèce \n\na,b\n"
    export = Mock(return_value=data)
    perenial_plants_list = Mock(return_value=Mock(export=export))
    web = Mock(perenial_plants_list=perenial_plants_list)

    plants = list(DEModel(web).get_perenial_plants())
    assert plants == [
        {
            "genus": "a",
            "species": "b",
        }
    ]


def test_de_ingestor_fetch_all():
    """Fetching all should return a list of elements."""
    model = Mock(
        get_perenial_plants=Mock(
            return_value=[
                {
                    "genus": "a",
                    "species": "b",
                    "english name": "c",
                    "french name": "d",
                }
            ]
        )
    )

    ingestor = DEIngestor(model)
    elements = list(ingestor.fetch_all())
    assert elements == [
        DatabasePlant(
            {
                "scientific name": "a b",
                "common name/c": True,
                "common name/d": True,
            }
        ),
    ]



"""Unit tests for the Natural Capital module."""

from textwrap import dedent
from unittest.mock import Mock, patch

import pytest
from yarl import URL

from permaculture.nc import (
    NCAuthentication,
    NCAuthenticationError,
    NCConverter,
    NCIngestor,
    NCLink,
    NCModel,
    NCWeb,
)

from .stubs import StubRequestsPreparedRequest, StubRequestsResponse


def test_nc_ingestor_from_config_password_file(tmpdir, unique):
    """Instantiating from config should read password from file."""
    password = unique("password")
    password_file = tmpdir / "secret"
    password_file.write_text(password, encoding="utf8")
    config = Mock(
        nc_username=unique("text"),
        nc_password=None,
        nc_password_file=str(password_file),
        storage=None,
    )
    with patch.object(NCModel, "with_authentication") as mock_auth:
        mock_auth.return_value = NCModel()
        NCIngestor.from_config(config)
        mock_auth.assert_called_once_with(
            config.nc_username,
            password,
            config.storage,
        )


def test_nc_ingestor_from_config_password(unique):
    """Instantiating from config should use password when no file."""
    password = unique("password")
    config = Mock(
        nc_username=unique("text"),
        nc_password=password,
        nc_password_file=None,
        storage=None,
    )
    with patch.object(NCModel, "with_authentication") as mock_auth:
        mock_auth.return_value = NCModel()
        NCIngestor.from_config(config)
        mock_auth.assert_called_once_with(
            config.nc_username,
            password,
            config.storage,
        )



def test_nc_authentication_get_payload(unique):
    username, password = unique("text"), unique("password")
    value, key = unique("text"), unique("text")
    response = StubRequestsResponse(
        text=dedent(
            f"""\
        <form name="btl-formlogin">
        <input name="bttask" value="login" />
        <input name="return" value="{value}" />
        <input name="{key}" value="1" />
        </form>
    """
        )
    )
    session = Mock(get=Mock(return_value=response))
    authentication = NCAuthentication(username, password, session)
    request = StubRequestsPreparedRequest()
    payload = authentication.get_payload(request)
    assert payload == {
        "bttask": "login",
        "username": username,
        "passwd": password,
        "return": value,
        key: "1",
        "remember": "yes",
    }


def test_nc_authentication_authenticate_without_cookies():
    """Authenticating without cookies in the response should raise."""
    response = StubRequestsResponse(headers={"Set-Cookie": ""})
    session = Mock(post=Mock(return_value=response))
    with patch.object(NCAuthentication, "get_payload", return_value={}):
        authentication = NCAuthentication(None, None, session)
        with pytest.raises(NCAuthenticationError):
            request = StubRequestsPreparedRequest()
            authentication.authenticate(request)


def test_nc_authentication_authenticate_with_cookies(unique):
    """Authenticating with cookies in the response should return a response."""
    cookies = unique("cookies")
    headers = {"Set-Cookie": "test"}
    response = StubRequestsResponse(headers=headers, cookies=cookies)
    session = Mock(post=Mock(return_value=response))
    with patch.object(NCAuthentication, "get_payload", return_value={}):
        authentication = NCAuthentication(None, None, session)
        request = StubRequestsPreparedRequest()
        cookie = authentication.authenticate(request)

    assert cookie == "test"


def test_nc_web_view_detail():
    """Viewing the detail for a plant should POST with the identifier."""
    session = Mock(post=Mock(return_value=StubRequestsResponse()))
    NCWeb(session).view_detail("i")
    session.post.assert_called_once_with(
        "/plant-database/new-the-plant-list",
        params={
            "vw": "detail",
            "id": "i",
        },
    )


def test_nc_web_view_list(unique):
    """Viewing the list of plants should POST with the plant names."""
    sci_name, sort_name = unique("text"), unique("text")
    session = Mock(post=Mock(return_value=StubRequestsResponse(text="test")))
    result = NCWeb(session).view_list(sci_name, sort_name)
    session.post.assert_called_once_with(
        "/plant-database/new-the-plant-list",
        params={
            "vw": "list",
        },
        data={
            "sortName": sort_name,
            "sciName": sci_name,
            "bfilter": "Set Filter",
            "limitstart": 0,
        },
    )
    assert result == "test"


@pytest.mark.parametrize(
    "item, expected",
    [
        pytest.param(
            ("Bloom Time", "Spring - Late Summer"),
            [
                ("bloom period/min", "spring"),
                ("bloom period/max", "summer"),
            ],
            id="Bloom Time",
        ),
        pytest.param(
            ("Soil Type", "Sandy, Loamy"),
            [
                ("soil type/sandy", True),
                ("soil type/loamy", True),
            ],
            id="Soil Type",
        ),
        pytest.param(
            ("Sun", "Full Sun, Partial Shade"),
            [
                ("sun/full", True),
                ("sun/partial", True),
            ],
            id="Sun",
        ),
        pytest.param(
            ("Height", "12 inches - 36-40 inches"),
            [("height/min", 0.30479999999999996), ("height/max", 0.9144)],
            id="Height",
        ),
    ],
)
def test_nc_converter_convert_item(item, expected):
    """Converting an item should consider types."""
    result = NCConverter().convert_item(*item)
    assert result == expected


@pytest.mark.parametrize(
    "text, expected",
    [
        pytest.param(
            dedent(
                """\
            <table width="100%">
              <tr>
                <td class="plantList">A</td>
              </tr>
            </table>",
        """
            ),
            [],
            id="0 rows",
        ),
        pytest.param(
            dedent(
                """\
            <table width="100%">
              <tr>
                <td class="plantList">A</td>
              </tr>
              <tr>
                <td>1</td>
              </tr>
            </table>",
        """
            ),
            [
                {"A": "1"},
            ],
            id="1 row",
        ),
        pytest.param(
            dedent(
                """\
            <table width="100%">
              <tr>
                <td class="plantList">A</td>
              </tr>
              <tr>
                <td><a href="http://example.com">1</a></td>
              </tr>
            </table>",
        """
            ),
            [
                {"A": NCLink("1", URL("http://example.com"))},
            ],
            id="1 link",
        ),
    ],
)
def test_nc_model_parse_plant_list(text, expected):
    """Parsing a plant list should return a list of dictionaries."""
    model = NCModel(None)
    result = list(model.parse_plant_list(text))
    assert result == expected


@pytest.mark.parametrize(
    "text, expected",
    [
        pytest.param(
            dedent(
                """\
            <table width="100%">
              <tr>
                <td><b>A</b> a</td>
              </tr>
            </table>",
        """
            ),
            {"A": "a"},
            id="1 row",
        ),
        pytest.param(
            dedent(
                """\
            <table width="100%">
              <tr>
                <td><b>A</b> a</td>
              </tr>
              <tr>
                <td><b>B</b> b</td>
              </tr>
            </table>",
        """
            ),
            {"A": "a", "B": "b"},
            id="2 rows",
        ),
        pytest.param(
            dedent(
                """\
            <table width="100%">
              <tr>
                <td><b>A</b> a</td>
              </tr>
              <tr>
                <td><b>B</b> b</td>
              </tr>
            </table>",
        """
            ),
            {"A": "a", "B": "b"},
            id="2 columns",
        ),
    ],
)
def test_nc_model_parse_detail(text, expected):
    """Parsing detail should return a dictionary."""
    model = NCModel(None)
    result = model.parse_detail(text)
    assert result == expected


@pytest.mark.parametrize(
    "text, expected",
    [
        pytest.param(
            dedent(
                """\
            <table width="100%">
              <tr>
                <td width="50%"><b>Items 1 - 50 of 100</td>
              </tr>
            </table>",
        """
            ),
            100,
            id="total",
        ),
    ],
)
def test_nc_model_parse_plant_total(text, expected):
    """Parsing a plant total should return a total number of plants."""
    model = NCModel(None)
    result = model.parse_plant_total(text)
    assert result == expected


def test_nc_model_get_plant_total():
    """Getting the plant total should the total number of plants."""
    with patch.object(NCWeb, "view_list") as mock_list:
        mock_list.return_value = dedent(
            """\
            <table width="100%">
            </table>
        """
        )
        web = NCWeb(None)
        model = NCModel(web)
        model.get_plant_total()
        mock_list.assert_called_once()



"""Unit tests for the priority module."""

from unittest.mock import Mock

from permaculture.priority import LocationPriority
from permaculture.storage import null_storage


def test_location_priority_weight(unique):
    """The location priority weight should be 1 over the distance."""
    distance = unique("float")
    multi_polygon = Mock(distance=Mock(return_value=distance))
    nominatim = Mock(multi_polygon=Mock(return_value=multi_polygon))
    priority = LocationPriority(None, nominatim, Mock())
    result = priority.weight
    assert result == 1 / distance


def test_location_priority_with_cache():
    """The location priority should cache nominatim and ipinfo."""
    nominatim = Mock()
    ipinfo = Mock()
    LocationPriority(None, nominatim, ipinfo).with_cache(null_storage)
    nominatim.with_cache.assert_called_once_with(null_storage)
    ipinfo.with_cache.assert_called_once_with(null_storage)



"""Unit test for the location module."""

import pytest

from permaculture.location import (
    LocationPoint,
    LocationPolygon,
)


def test_location_point_sequence():
    """A point should behave like a sequence of lat and lon."""
    result = LocationPoint(0, 1)
    assert tuple(result) == (0, 1)


@pytest.mark.parametrize(
    "point, expected",
    [
        (LocationPoint(0, 0), 0),
        (LocationPoint(1, 0), 1),
        (LocationPoint(0, 1), 1),
        (LocationPoint(1, 1), 2),
    ],
)
def test_location_point_normsqr(point, expected):
    """The euclidean norm squared should use the latitude and longitude."""
    result = point.normsqr
    assert result == expected


@pytest.mark.parametrize(
    "point, other, expected",
    [
        (LocationPoint(0, 0), (0, 0), 0),
        (LocationPoint(0, 0), (1, 0), 111),
        (LocationPoint(0, 0), (0, 1), 111),
        (LocationPoint(0, 0), (1, 1), 157),
    ],
)
def test_location_point_distance(point, other, expected):
    """The distance between two points return the great-circle distance."""
    result = point.distance(other)
    assert int(result) == expected


@pytest.mark.parametrize(
    "point, other, expected",
    [
        (LocationPoint(0, 0), (0, 0), 0),
        (LocationPoint(1, 0), (1, 0), 1),
        (LocationPoint(0, 1), (0, 1), 1),
        (LocationPoint(1, 1), (1, 1), 2),
    ],
)
def test_location_point_dot(point, other, expected):
    """The dot product should use the latitude and longituide."""
    result = point.dot(other)
    assert result == expected


@pytest.mark.parametrize(
    "point, other, expected",
    [
        (LocationPoint(0, 0), (0, 0), (0, 0)),
        (LocationPoint(1, 0), (0, 1), (1, 1)),
        (LocationPoint(0, 1), (1, 0), (1, 1)),
        (LocationPoint(1, 1), (1, 1), (2, 2)),
    ],
)
def test_location_point_add(point, other, expected):
    """Adding two points should return the resulting point."""
    result = point + other
    assert result == expected


@pytest.mark.parametrize(
    "point, other, expected",
    [
        (LocationPoint(0, 0), (0, 0), (0, 0)),
        (LocationPoint(1, 0), (0, 1), (1, -1)),
        (LocationPoint(0, 1), (1, 0), (-1, 1)),
        (LocationPoint(1, 1), (1, 1), (0, 0)),
    ],
)
def test_location_point_sub(point, other, expected):
    """Subtracting two points should return the resulting point."""
    result = point - other
    assert result == expected


@pytest.mark.parametrize(
    "point, other, expected",
    [
        (LocationPoint(0, 0), 2, (0, 0)),
        (LocationPoint(1, 0), 2, (2, 0)),
        (LocationPoint(0, 1), 2, (0, 2)),
        (LocationPoint(1, 1), 2, (2, 2)),
    ],
)
def test_location_point_mul(point, other, expected):
    """Multiplying a point by a number should return the resulting point."""
    result = point * other
    assert result == expected


@pytest.mark.parametrize(
    "point, other, expected",
    [
        (LocationPoint(0, 0), 2, (0, 0)),
        (LocationPoint(2, 0), 2, (1, 0)),
        (LocationPoint(0, 2), 2, (0, 1)),
        (LocationPoint(2, 2), 2, (1, 1)),
    ],
)
def test_location_point_div(point, other, expected):
    """Dividing a point by a number should return the resulting point."""
    result = point / other
    assert result == expected


@pytest.mark.parametrize(
    "polygon, expected",
    [
        (
            LocationPolygon.from_points(
                [[0, 0], [0, 1], [1, 1], [1, 0], [0, 0]],
            ),
            1,
        ),
        (
            LocationPolygon.from_points(
                [[1, 6], [3, 1], [7, 2], [4, 4], [8, 5]],
            ),
            16.5,
        ),
        (
            LocationPolygon.from_points(
                [[1, 6], [3, 1], [7, 2], [4, 4], [8, 5], [1, 6]],
            ),
            16.5,
        ),
    ],
)
def test_location_polygon_area(polygon, expected):
    """The area should return the area of a simple polygon in a plane."""
    result = polygon.area
    assert result == expected


@pytest.mark.parametrize(
    "polygon, point, expected",
    [
        (
            LocationPolygon.from_points(
                [[0, 0], [0, 1], [1, 1], [1, 0]],
            ),
            (0, 0),
            True,
        ),
        (
            LocationPolygon.from_points(
                [[20, 10], [50, 125], [125, 90], [150, 10]],
            ),
            (75, 50),
            True,
        ),
        (
            LocationPolygon.from_points(
                [[20, 10], [50, 125], [125, 90], [150, 10]],
            ),
            (200, 50),
            False,
        ),
        (
            LocationPolygon.from_points(
                [[20, 10], [50, 125], [125, 90], [150, 10]],
            ),
            (35, 90),
            False,
        ),
        (
            LocationPolygon.from_points(
                [[20, 10], [50, 125], [125, 90], [150, 10]],
            ),
            (50, 10),
            True,
        ),
    ],
)
def test_location_polygon_contains(polygon, point, expected):
    """The polygon contains a point when it's inside the edges."""
    result = polygon.contains(point)
    assert result == expected


@pytest.mark.parametrize(
    "polygon, point, expected",
    [
        (LocationPolygon.from_points([[0, 0]]), (0, 0), (0, 0)),
        (LocationPolygon.from_points([[1, 1]]), (0, 0), (1, 1)),
        (LocationPolygon.from_points([[1, 1], [0, 0]]), (0, 0), (0, 0)),
        (LocationPolygon.from_points([[0, 0], [2, 0]]), (1, 1), (1, 0)),
    ],
)
def test_location_polygon_nearest(polygon, point, expected):
    """The nearest point to a polygon should be the expected point."""
    result = polygon.nearest(point)
    assert result == expected



"""Unit tests for the unit module."""

import math

import pytest

from permaculture.unit import (
    celsius,
    fahrenheit,
    feet,
    inches,
    km,
    meters,
    miles,
)


@pytest.mark.parametrize(
    "a, b",
    [
        pytest.param(1 * feet, 0.3048 * meters, id="feet to meters"),
        pytest.param(1 * meters, 3.28084 * feet, id="meters to feet"),
        pytest.param(1 * feet, 12 * inches, id="feet to inches"),
        pytest.param(
            100 * celsius, 212 * fahrenheit, id="celsius to fahrenheit"
        ),
        pytest.param(1 * km, 1000 * meters, id="km to meters"),
        pytest.param(1 * km, 0.621371 * miles, id="km to miles"),
    ],
)
def test_unit_rmul(a, b):
    """It should be convenient to convert between units."""
    assert math.isclose(a, b, rel_tol=1e-5, abs_tol=0.0)



"""Unit tests for the logger module."""
import json
import logging
import sys
from argparse import ArgumentParser
from contextlib import contextmanager
from io import StringIO

import pytest
from hamcrest import (
    assert_that,
    has_entries,
    has_properties,
    is_,
    matches_regexp,
)

from permaculture.logger import (
    JsonFormatter,
    LoggerFormatter,
    LoggerHandlerAction,
    LoggerLevelAction,
    logger_context,
    remove_log_context,
    set_log_context,
    setup_logger,
)

DATE_PATTERN = r"\d{4}\-\d{2}\-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}"


@contextmanager
def logger_stream(formatter):
    """Context manager for a stream used by a logger handler at debug level."""
    stream = StringIO()
    handler = logging.StreamHandler(stream)

    setup_logger(logging.DEBUG, handler, formatter)
    try:
        yield stream
    finally:
        setup_logger()


def test_logger_formatter_default():
    """The LoggerFormatter should include microseconds by default."""
    with logger_stream(LoggerFormatter()) as stream:
        message = "message"
        logging.debug(message)
        stream.seek(0)

        assert_that(
            stream.readline(),
            matches_regexp(f'{DATE_PATTERN} {"DEBUG":<7} root {message}\n'),
        )


def test_logger_formatter_non_root_logger():
    """The LoggerFormatter includes the name even for non-root loggers."""
    with logger_stream(LoggerFormatter()) as stream:
        message = "message"
        logger = logging.getLogger("deeply.nested.logger")
        logger.error(message)
        stream.seek(0)

        assert_that(
            stream.readline(),
            matches_regexp(
                f'{DATE_PATTERN} {"ERROR":<7} deeply.nested.logger {message}\n'
            ),
        )


def test_logger_formatter_custom_fmt():
    """The LoggerFormatter can take fmt as argument."""
    with logger_stream(LoggerFormatter(fmt="%(message)s")) as stream:
        message = "message"
        logging.info(message)
        stream.seek(0)

        assert_that(
            stream.readline(),
            f"{message}\n",
        )


def test_logger_formatter_custom_datefmt():
    """The LoggerFormatter can also take datefmt as argument."""
    with logger_stream(
        LoggerFormatter(datefmt="%a, %d %b %Y %H:%M:%S")
    ) as stream:
        date_pattern = r"\w{3}, \d{2} \w{3} \d{4} \d{2}:\d{2}:\d{2}"
        message = "message"
        logging.info(message)
        stream.seek(0)

        assert_that(
            stream.readline(),
            matches_regexp(f'{date_pattern} {"INFO":<7} root {message}\n'),
        )


def test_json_formatter_default():
    """The JsonFormatter should include Level, Message, Name and Timestamp."""
    with logger_stream(JsonFormatter()) as stream:
        message = "message"
        logging.debug(message)
        stream.seek(0)

        assert_that(
            json.loads(stream.readline()),
            has_entries(
                Level="DEBUG",
                Message=message,
                Name="root",
                Timestamp=matches_regexp(DATE_PATTERN),
            ),
        )


def test_json_formatter_context():
    """The JsonFormatter should include context data."""
    with logger_stream(JsonFormatter()) as stream:
        message = "message"
        source = "source"
        with logger_context({"Source": source}):
            logging.debug(message)
        stream.seek(0)

        assert_that(
            json.loads(stream.readline()),
            has_entries(
                Level="DEBUG",
                Message=message,
                Name="root",
                Timestamp=matches_regexp(DATE_PATTERN),
                Source=source,
            ),
        )


def test_logger_handler_action_default():
    """A LoggerHandlerAction should default to stderr."""
    parser = ArgumentParser()
    parser.add_argument("--handler", action=LoggerHandlerAction)

    result = parser.parse_args([])

    assert_that(
        result,
        has_properties(
            handler=has_properties(
                stream=sys.stderr,
                formatter=is_(LoggerFormatter),
            )
        ),
    )


def test_logger_handler_action_stdout():
    """A LoggerHandlerAction should use stdout with -."""
    parser = ArgumentParser()
    parser.add_argument("--handler", action=LoggerHandlerAction)

    result = parser.parse_args(["--handler", "-"])

    assert_that(
        result,
        has_properties(
            handler=has_properties(
                stream=sys.stdout,
                formatter=is_(LoggerFormatter),
            )
        ),
    )


def test_logger_handler_action_file(tmpdir):
    """A LoggerHandlerAction should use file stream with a filename."""
    parser = ArgumentParser()
    parser.add_argument("--handler", action=LoggerHandlerAction)
    filename = str(tmpdir.join("test.log"))

    result = parser.parse_args(["--handler", filename])

    assert_that(
        result,
        has_properties(
            handler=has_properties(
                baseFilename=filename,
                formatter=is_(LoggerFormatter),
            )
        ),
    )


def test_logger_level_action_default():
    """A LoggerLevelAction should default to INFO."""
    parser = ArgumentParser()
    parser.add_argument("--level", action=LoggerLevelAction)

    result = parser.parse_args([])

    assert_that(result, has_properties(level=logging.INFO))


@pytest.mark.parametrize(
    "choice, level",
    [
        ("debug", logging.DEBUG),
        ("Debug", logging.DEBUG),
        ("DEBUG", logging.DEBUG),
    ],
)
def test_logger_level_action_valid_choice(choice, level):
    """A LoggerLevelAction should set on a valid choice."""
    parser = ArgumentParser()
    parser.add_argument("--level", action=LoggerLevelAction)

    result = parser.parse_args(["--level", choice])

    assert_that(result, has_properties(level=level))


def test_logger_level_action_invalid_choice():
    """A LoggerLevelAction should exit on an invalid choice."""
    parser = ArgumentParser()
    parser.add_argument("--level", action=LoggerLevelAction)

    with pytest.raises(SystemExit):
        parser.parse_args(["--level", "test"])


def test_logger_level_action_unsupported_choice():
    """A LoggerLevelAction should exit on an unsupported choice."""
    parser = ArgumentParser()
    parser.add_argument("--level", action=LoggerLevelAction, choices=["test"])

    with pytest.raises(SystemExit):
        parser.parse_args(["--level", "test"])


@pytest.mark.parametrize(
    "level, records",
    [
        pytest.param(
            logging.ERROR,
            [
                has_properties(levelname="ERROR"),
            ],
            id="ERROR",
        ),
        pytest.param(
            logging.WARNING,
            [
                has_properties(levelname="ERROR"),
                has_properties(levelname="WARNING"),
            ],
            id="WARNING",
        ),
        pytest.param(
            logging.INFO,
            [
                has_properties(levelname="ERROR"),
                has_properties(levelname="WARNING"),
                has_properties(levelname="INFO"),
            ],
            id="INFO",
        ),
        pytest.param(
            logging.DEBUG,
            [
                has_properties(levelname="ERROR"),
                has_properties(levelname="WARNING"),
                has_properties(levelname="INFO"),
                has_properties(levelname="DEBUG"),
            ],
            id="DEBUG",
        ),
    ],
)
def test_setup_logger_level(level, records, logger_handler):
    """The log level should output expected records."""
    logger = setup_logger(level, logger_handler)
    logger.error("This is error")
    logger.warning("This is warning")
    logger.info("This is info")
    logger.debug("This is debug")

    result = logger_handler.records

    assert_that(result, records)


def test_setup_logger_formatter(logger_handler):
    message = "message"
    logger = setup_logger(logging.DEBUG, logger_handler, JsonFormatter())
    logger.debug(message)

    record = logger_handler.records[0]
    result = json.loads(logger_handler.format(record))

    assert_that(
        result,
        has_entries(
            Level="DEBUG",
            Message=message,
            Name="root",
            Timestamp=matches_regexp(DATE_PATTERN),
        ),
    )


def test_set_remove_log_context(logger_handler):
    """Logging context is set and removed from log entries."""
    logging.info("Test log0")
    set_log_context({"a": "1"})
    logging.info("Test log1")
    set_log_context({"b": "2"})
    logging.info("Test log2")

    # Remove the contexts
    remove_log_context("a")
    logging.info("Test log3")
    remove_log_context("b")
    logging.info("Test log4")

    assert_that(
        logger_handler.records,
        [
            has_properties(msg="Test log0"),
            has_properties(msg='[{"a": "1"}] Test log1'),
            has_properties(msg='[{"a": "1", "b": "2"}] Test log2'),
            has_properties(msg='[{"b": "2"}] Test log3'),
            has_properties(msg="Test log4"),
        ],
    )


def test_logger_context(logger_handler):
    """Logging contexts are added to log lines within the context only."""
    logging.info("Test log0")

    with logger_context({"a": "1"}):
        logging.info("Test log1")
        with logger_context({"b": "2"}):
            logging.info("Test log2")
            with logger_context({"a": "3"}):
                logging.info("Test log3")
        logging.info("Test log4")
    logging.info("Test log5")
    with logger_context({"a": "1", "b": "2"}):
        logging.info("Test log6")

    assert_that(
        logger_handler.records,
        [
            has_properties(msg="Test log0"),
            has_properties(msg='[{"a": "1"}] Test log1'),
            has_properties(msg='[{"a": "1", "b": "2"}] Test log2'),
            has_properties(msg='[{"a": "3", "b": "2"}] Test log3'),
            has_properties(msg='[{"a": "1"}] Test log4'),
            has_properties(msg="Test log5"),
            has_properties(msg='[{"a": "1", "b": "2"}] Test log6'),
        ],
    )



"""Unit tests for the pfaf module."""

from pathlib import Path
from unittest.mock import Mock

import pytest

from permaculture.database import DatabasePlant
from permaculture.pfaf import (
    PFAFConverter,
    PFAFFile,
    PFAFIngestor,
    PFAFModel,
)


def test_pfaf_file_main_database():
    """The main database should return the corresponding Excel worksheet."""
    path = Path(__file__).with_suffix(".xls")
    file = PFAFFile(path)
    ws = file.main_database()
    assert ws.cell(0, 0).value == "Test"


def test_pfaf_file_main_database_error(tmp_path):
    """The main database should raise when file not found."""
    file = PFAFFile(tmp_path / "nonexistent.xls")
    with pytest.raises(FileNotFoundError):
        file.main_database()


@pytest.mark.parametrize(
    "value, expected",
    [
        pytest.param("1.0", 1.0, id="str"),
        pytest.param(1.0, 1.0, id="float"),
        pytest.param(1, 1.0, id="int"),
    ],
)
def test_pfaf_converter_convert_float(value, expected):
    """Converting a float should accept a float or a string."""
    result = PFAFConverter().convert_float("", value)
    assert result == [("", expected)]


@pytest.mark.parametrize(
    "item, expected",
    [
        pytest.param(
            ("Deciduous/Evergreen", "DE"),
            [
                ("deciduous/evergreen/deciduous", True),
                ("deciduous/evergreen/evergreen", True),
            ],
            id="Deciduous/Evergreen",
        ),
        pytest.param(
            ("Growth rate", "S"),
            [("growth rate/slow", True)],
            id="Growth rate slow",
        ),
        pytest.param(
            ("Growth rate", "M"),
            [("growth rate/medium", True)],
            id="Growth rate medium",
        ),
        pytest.param(
            ("Growth rate", "F"),
            [("growth rate/fast", True)],
            id="Growth rate fast",
        ),
        pytest.param(
            ("Moisture", "DMWeWa"),
            [
                ("moisture/dry", True),
                ("moisture/moist", True),
                ("moisture/wet", True),
                ("moisture/water", True),
            ],
            id="Moisture",
        ),
        pytest.param(
            ("pH", "ANB"),
            [
                ("ph/acid", True),
                ("ph/neutral", True),
                ("ph/alkaline", True),
            ],
            id="pH",
        ),
        pytest.param(
            ("Shade", "FSN"),
            [
                ("sun/shade", True),
                ("sun/partial", True),
                ("sun/full", True),
            ],
            id="Shade",
        ),
        pytest.param(
            ("Soil", "LMH"),
            [
                ("soil/light", True),
                ("soil/medium", True),
                ("soil/heavy", True),
            ],
            id="Soil",
        ),
    ],
)
def test_pfaf_converter_convert_item(item, expected):
    """Converting an item should consider types."""
    result = PFAFConverter().convert_item(*item)
    assert result == expected


def test_pfaf_model_all_plants_error(tmp_path):
    """All plants should return no plant when file not found."""
    model = PFAFModel.from_path(tmp_path / "nonexistent.xls")
    plants = list(model.all_plants())
    assert plants == []


def test_pfaf_ingestor_fetch_all():
    """Fetching all should return a list of elements."""
    model = Mock(
        all_plants=Mock(
            return_value=[
                {
                    "scientific name": "a",
                    "common name": "b",
                }
            ]
        )
    )

    ingestor = PFAFIngestor(model)
    elements = list(ingestor.fetch_all())
    assert elements == [
        DatabasePlant(
            {
                "scientific name": "a",
                "common name": "b",
            }
        ),
    ]



"""Unit tests for the cli module."""

import logging
from unittest.mock import patch

import pytest

from permaculture.cli import (
    main,
    make_args_parser,
    make_config_parser,
)


def test_make_args_parser_command(unique):
    """Making an args parser should parse known commands."""
    args_parser = make_args_parser()
    args, _ = args_parser.parse_known_args(["lookup", unique("text")])
    assert args.command == "lookup"


def test_make_args_parser_ingest():
    """Making an args parser should parse the ingest command."""
    args_parser = make_args_parser()
    args, _ = args_parser.parse_known_args(["ingest"])
    assert args.command == "ingest"


def test_make_args_parser_command_error(unique):
    """Making an args parser should raise on unknown commands."""
    args_parser = make_args_parser()
    with pytest.raises(SystemExit):
        args_parser.parse_known_args(["test"])


def test_make_args_parser_remaining(unique):
    """Making an args parser should have remaining args."""
    args_parser = make_args_parser()
    _, remaining = args_parser.parse_known_args(
        ["--log-level=debug", "lookup", unique("text")]
    )
    assert remaining == ["--log-level=debug"]


def test_make_config_parser_defaults():
    """Making a config parser should have sensible defaults."""
    config_parser = make_config_parser([])
    config = config_parser.parse_args([])
    assert config.log_level == logging.WARNING


def test_make_config_parser_args():
    """Making a config parser should parse args."""
    config_parser = make_config_parser([])
    config = config_parser.parse_args(["--log-level=debug"])
    assert config.log_level == logging.DEBUG


def test_make_config_parser_files(tmpdir):
    """Making a config parser should parse files."""
    path = tmpdir / ".permaculture"
    path.write_text("log-level = debug", encoding="utf8")
    config_parser = make_config_parser([path])
    config = config_parser.parse_args([])
    assert config.log_level == logging.DEBUG


def test_make_config_parser_env(monkeypatch):
    """Making a config parser should read PERMACULTURE_ env vars."""
    monkeypatch.setenv("PERMACULTURE_NC_PASSWORD", "secret")
    config_parser = make_config_parser([])
    config = config_parser.parse_args([])
    assert config.nc_password == "secret"  # noqa: S105


def test_make_config_parser_nc_password_file(tmpdir):
    """Making a config parser should accept --nc-password-file."""
    config_parser = make_config_parser([])
    password_file = tmpdir / "secret"
    password_file.write_text("secret", encoding="utf8")
    config = config_parser.parse_args([f"--nc-password-file={password_file}"])
    assert config.nc_password_file == str(password_file)


@patch("sys.stdout")
def test_main_help(stdout):
    """The main function should output usage when asked for --help."""
    with pytest.raises(SystemExit):
        main(["--help"])

    assert "usage" in stdout.write.call_args[0][0]


@patch("sys.stderr")
def test_main_error(stderr, unique):
    """The main function should output an error with an invalid command."""
    with pytest.raises(SystemExit):
        main([unique("text")])

    assert "invalid choice" in stderr.write.call_args[0][0]



"""Google spreadsheets web interface."""

from pathlib import Path

from attrs import define
from yarl import URL

from permaculture.http import HTTPSession
from permaculture.storage import null_storage


@define(frozen=True)
class GoogleSpreadsheet:
    """Google spreadsheet."""

    session: HTTPSession
    doc_id: str

    @classmethod
    def from_url(cls, url: URL, storage=null_storage):
        session = HTTPSession(url.origin()).with_cache(storage)
        doc_id = Path(url.path).parent.name
        return cls(session, doc_id)

    def export(self, gid=1, fmt="csv"):
        response = self.session.get(
            f"/spreadsheets/d/{self.doc_id}/export",
            params={
                "gid": gid,
                "format": fmt,
            },
        )

        return response.content.decode("utf8")



"""Unit testing stubs."""

from collections.abc import Callable
from http.cookiejar import CookieJar

from attrs import Factory, define, field


@define(frozen=True)
class StubRequestsPreparedRequest:
    """Stub Requests PreparedRequest object."""

    method: str = "GET"
    headers: dict = field(factory=dict)
    body: str = ""
    url: str = "http://www.test.com/"

    def prepare_cookies(self, _):
        """Do nothing."""


@define(frozen=True)
class StubRequestsResponse:
    """Stub Requests Response object."""

    status_code: int = 200
    headers: dict = field(factory=dict)
    json: Callable[[], dict] = lambda: {}
    url: str = "http://www.test.com/"
    reason: str = ""
    text: str = ""
    cookies: CookieJar = None
    request: StubRequestsPreparedRequest = field(
        default=Factory(
            lambda self: StubRequestsPreparedRequest(url=self.url),
            takes_self=True,
        )
    )



"""Action utilities."""

from argparse import Action


def enum_action(enum):
    """Return an argparse action class for an Enum type."""

    class EnumAction(Action):
        """Argparse action for an Enum type."""

        def __init__(self, option_strings, dest, **kwargs):
            """Initialize enum defaults."""
            kwargs.setdefault("choices", [t.name for t in enum])
            super().__init__(option_strings, dest, **kwargs)

        def __call__(self, parser, namespace, values, option_string=None):
            """Set the dest attribute to the Enum type."""
            if isinstance(values, list):
                values = [enum[v] for v in values]
            else:
                values = enum[values]

            setattr(namespace, self.dest, values)

    return EnumAction


class SingleAction(Action):
    """Argparse action where nargs is not allowed."""

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        """Raise a ValueError if nargs is set."""
        if nargs is not None:
            raise ValueError("nargs not allowed")

        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        """Set the dest attribute in the namespace."""
        setattr(namespace, self.dest, values)



"""Unit tests for the storage module."""

from argparse import ArgumentParser
from pathlib import Path

import pytest
from hamcrest import assert_that, has_properties, is_

from permaculture.storage import (
    FileStorage,
    MemoryStorage,
    SqliteStorage,
    StorageAction,
    null_storage,
)


@pytest.fixture(
    params=[
        "file",
        "memory",
        "sqlite",
    ]
)
def real_storage(request, tmpdir):
    """Produce pytest parameters for all storages."""
    if request.param == "file":
        yield FileStorage(tmpdir)
    elif request.param == "memory":
        yield MemoryStorage()
    elif request.param == "sqlite":
        yield SqliteStorage.from_path(tmpdir / "storage.sqlite")
    else:
        raise KeyError(f"Unsupported storage type: {request.param}")


@pytest.fixture(
    params=[
        "a",
        "a/b",
        "http://a/b",
    ]
)
def key(request):
    """Produce a key to test the storage."""
    return request.param


def test_storage_action_default():
    """A SerializerAction should default to memory."""
    parser = ArgumentParser()
    parser.add_argument("--storage", action=StorageAction)
    result = parser.parse_args([])

    assert_that(
        result,
        has_properties(storage=is_(MemoryStorage)),
    )


def test_storage_action_custom():
    """A StorageAction should use the path given as argument."""
    parser = ArgumentParser()
    parser.add_argument("--storage", action=StorageAction)
    result = parser.parse_args(["--storage", "/path"])

    assert_that(
        result,
        has_properties(storage=has_properties(base_dir=Path("/path"))),
    )


def test_real_storage_getitem_non_existing(key, real_storage):
    """Getting a non-existing key should return the default value."""
    assert real_storage.get(key, True)


def test_real_storage_getitem_existing(key, real_storage):
    """Getting a stored key should only return its value."""
    real_storage[key] = True
    assert real_storage.get(key, False)


def test_real_storage_delitem_non_existing(key, real_storage):
    """Deleting a non-existing key should raise a KeyError."""
    with pytest.raises(KeyError):
        del real_storage[key]


def test_real_storage_delitem_existing(key, real_storage):
    """Deleting a stored key should then return the default value."""
    real_storage[key] = True
    del real_storage[key]
    assert real_storage.get(key, True)


def test_real_storage_iter(key, real_storage):
    """Iterating over a storage should return the keys."""
    real_storage[key] = True
    assert list(real_storage) == [key]


def test_real_storage_length(key, real_storage):
    """The length should return the number of stored keys."""
    assert len(real_storage) == 0
    real_storage[key] = True
    assert len(real_storage) == 1


def test_null_storage_getitem(key):
    """Getting an existing key should always return the default."""
    null_storage[key] = True
    assert not null_storage.get(key, False)


def test_null_storage_setitem(key):
    """Getting an existing key should always return the default."""
    null_storage[key] = True
    assert not null_storage.get(key, False)


def test_null_storage_delitem(key):
    """Deleting a key should always raise."""
    null_storage[key] = True
    with pytest.raises(KeyError):
        del null_storage[key]


def test_null_storage_iter(key):
    """Iterating over a null storage should return an empty list."""
    null_storage[key] = True
    assert not list(null_storage)


def test_null_storage_length(key):
    """The length of a null storage should always be 0."""
    null_storage[key] = True
    assert len(null_storage) == 0


def test_file_storage_setitem(key, tmpdir):
    """Setting a key should create the parent directory."""
    parent = tmpdir / "parent"
    assert not parent.exists()

    storage = FileStorage(parent)
    storage[key] = True
    assert parent.exists()


def test_sqlite_storage_from_path_twice(key, tmpdir):
    """Instantiating an SQLite storage from path twice should not fail."""
    path = tmpdir / "storage.sqlite"
    SqliteStorage.from_path(path)
    SqliteStorage.from_path(path)



"""Unit tests for the nlp module."""

from argparse import ArgumentTypeError

import pytest

from permaculture.nlp import Extractor, normalize, score, score_type


def scorer(a, b):
    """Normalize the distance between 0 and 1 for two letters."""
    return 1 - abs(ord(a) - ord(b)) / 26


@pytest.mark.parametrize(
    "text, expected",
    [
        pytest.param("foo", "foo", id="one word"),
        pytest.param("foô", "foo", id="accents"),
        pytest.param(" foo", "foo", id="leading space"),
        pytest.param("foo ", "foo", id="trailing space"),
        pytest.param("foo bar", "foo bar", id="two words"),
        pytest.param("foo  bar", "foo bar", id="space between"),
        pytest.param("foo x bar", "foo bar", id="single letter"),
        pytest.param("foo 012 bar", "foo bar", id="digits"),
        pytest.param("foo () baz", "foo baz", id="empty parenthesis"),
        pytest.param("foo (bar) baz", "foo baz", id="one set of parentheses"),
        pytest.param("(foo) bar (baz)", "bar", id="two sets of parentheses"),
        pytest.param("foo [] baz", "foo baz", id="empty brackets"),
        pytest.param("foo [bar] baz", "foo baz", id="one set of brackets"),
        pytest.param("[foo] bar [baz]", "bar", id="two sets of brackets"),
        pytest.param("FOO", "foo", id="lower case"),
        pytest.param("-foo-bar-", "foo bar", id="dash"),
        pytest.param(".foo..bar.", "foo bar", id="dot"),
    ],
)
def test_normalize(text, expected):
    """Normalizing text should return the expected result."""
    result = normalize(text)
    assert result == expected


@pytest.mark.parametrize(
    "a, b, expected",
    [
        ("a", "a", 1.0),
        ("ab", "ab", 1.0),
        ("a a", "a a", 1.0),
        ("a b", "a b", 1.0),
        ("a b", "b a", 1.0),
        ("a", "b", 0.0),
    ],
)
def test_score(a, b, expected):
    """Scoring two strings should return the Levenshtein distance."""
    result = score(a, b)
    assert result == expected


@pytest.mark.parametrize(
    "score, expected",
    [
        ("0", 0.0),
        ("0.5", 0.5),
        ("1.0", 1.0),
    ],
)
def test_score_type(score, expected):
    """A valid score type should return a float."""
    result = score_type(score)
    assert result == expected


@pytest.mark.parametrize(
    "score",
    [
        "a",
        "2",
        "-1",
    ],
)
def test_score_type_error(score):
    """An invalid score type should raise."""
    with pytest.raises(ArgumentTypeError):
        score_type(score)


@pytest.mark.parametrize(
    "query, choices, expected",
    [
        ("a", [], []),
        ("a", ["a", "b"], [(1, "a"), (1, "b")]),
    ],
)
def test_extractor_extract(query, choices, expected):
    """Extracting should return the scores for each choice."""
    result = Extractor(query).extract(choices)
    assert list(result) == expected


@pytest.mark.parametrize(
    "query, choices, expected",
    [
        ("a", ["a", "b"], (1, "a")),
        ("b", ["a", "b"], (1, "b")),
    ],
)
def test_extractor_extract_one(query, choices, expected):
    """Extracting one should return the choice with the best score."""
    result = Extractor(query, scorer=scorer).extract_one(choices)
    assert result == expected


def test_extractor_extract_one_error():
    """Extracting one result from empty choices should raise."""
    with pytest.raises(ValueError):
        Extractor("a").extract_one([])


@pytest.mark.parametrize(
    "query, choices, expected",
    [
        ("a", ["a", "b"], "a"),
        ("b", ["a", "b"], "b"),
        ("a", ["b", "c"], "b"),
    ],
)
def test_extractor_choose(query, choices, expected):
    """Extract the choice with the best score."""
    result = Extractor(query, scorer=scorer).choose(choices)
    assert result == expected


def test_extractor_choose_error():
    """Extract a choice from empty choices should raise."""
    with pytest.raises(ValueError):
        Extractor("a").choose([])



"""Data manipulation functions.

Flattening nested data into flat data:

    >>> flatten({'a': {'b': 1}})
    {'a/b': 1}

Unflattening flat data into nested data:

    >>> unflatten({'a/b': 1})
    {'a': {'b': 1}}

Merging a list of dictionaries by calculating the mean on collision:

    >>> merge([{'a': 1}, {'a': 2}], lambda _, v: sum(v)/len(v))
    {'a': 1.5}
"""

from collections.abc import Mapping
from itertools import groupby
from operator import itemgetter
from statistics import mean


def flatten(data, sep="/"):
    """Flatten nested data into flat data."""

    def _flatten(d, parent=None):
        for key, value in d:
            new_parent = f"{key}" if parent is None else f"{parent}{sep}{key}"
            if isinstance(value, Mapping):
                yield from _flatten(value.items(), new_parent)
            elif isinstance(value, list):
                yield from _flatten(enumerate(value), new_parent)
            else:
                yield new_parent, value

    d = data.items() if isinstance(data, Mapping) else enumerate(data)
    return dict(_flatten(d))


def unflatten(data, sep="/"):
    """Unflatten flat data into nested data."""
    d = None
    for key, value in data.items():
        tokens = key.split(sep)
        if d is None:
            s = d = [] if tokens[0].isdigit() else {}
        else:
            s = d

        indexed_tokens = zip(tokens, [*tokens[1:], value], strict=True)
        for count, (index, next_token) in enumerate(indexed_tokens, 1):
            if count == len(tokens):
                value = next_token
            elif next_token.isdigit():
                value = []
            else:
                value = {}

            if isinstance(s, list):
                index = int(index)
                while index >= len(s):
                    s.append(value)
            elif isinstance(s, dict):
                if index not in s:
                    s[index] = value
            else:
                raise TypeError(f"Conflicting types for {key!r} in {data}")

            s = s[index]

    return d


def resolve(key, values):
    """Default function to resolve collisions when merging dictionaries."""
    if isinstance(values[0], str):
        values = [v for v in values if v != ""]
        return values[0] if len(values) == 1 else values
    elif isinstance(values[0], float | int):
        return mean(values)
    else:
        raise TypeError(f"Unsupported values for {key!r}: {values!r}")


def merge(dicts, resolve=resolve):
    """Merge a list of dictionaries by resolving collisions.

    :param resolve: Function called on collision to resolve colliding values.
    """
    d = {}

    for key, values in groupby(
        sorted((item for d in dicts for item in d.items()), key=itemgetter(0)),
        itemgetter(0),
    ):
        values = list(map(itemgetter(1), values))
        if len(set(map(type, values))) > 1:
            raise ValueError(f"Different types for {key!r}: {values!r}")

        if isinstance(values[0], Mapping):
            d[key] = merge(values, resolve)
        elif isinstance(values[0], list):
            d[key] = list({i for v in values for i in v})
        elif len(values) == 1 or len(set(values)) == 1:
            d[key] = values[0]
        else:
            d[key] = resolve(key, values)

    return d



"""Integration tests for the testing logger module."""

import logging

from hamcrest import assert_that, has_properties

from permaculture.testing.logger import logger_time


def test_logger_time_default(logger_handler):
    """The default time in the context manager is 0.0."""
    with logger_time():
        logging.info("foo")

    assert_that(
        logger_handler.records,
        [
            has_properties(created=0),
        ],
    )


def test_logger_time_seconds(logger_handler):
    """The time can be passed in seconds."""
    seconds = 1.1
    with logger_time(seconds):
        logging.info("foo")

    assert_that(
        logger_handler.records,
        [
            has_properties(created=seconds),
        ],
    )



"""Serializers for web related formats."""
import csv
import json
import pickle
from collections.abc import Callable
from io import StringIO
from typing import Any
from urllib.parse import parse_qsl, urlencode

import yaml
from attrs import define, field

from permaculture.action import SingleAction
from permaculture.registry import registry_load


class SerializerNotFound(Exception):
    """Raised when a serializer is not found for a content type."""


class SerializerAction(SingleAction):
    """Argument action for a serializer."""

    metavar = "CONTENT-TYPE"

    def __init__(self, option_strings, registry=None, **kwargs):
        """Initializer serializer defaults."""
        default = kwargs.pop("default", "application/x-yaml")
        kwargs.setdefault("default", self.get_serializer(default))
        kwargs.setdefault("metavar", self.metavar)
        kwargs.setdefault(
            "help", f"serializer content-type (default {default})"
        )
        super().__init__(option_strings, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        """Set the values to a serializer."""
        serializer = self.get_serializer(values)

        super().__call__(parser, namespace, serializer, option_string)

    @classmethod
    def get_serializer(cls, content_type):
        """Get a serializer with a default content-type."""
        return Serializer.load(content_type)


@define(frozen=True)
class Serializer:
    """A serializer to encode and decode data.

    :param content_type: Default content type of the serializer, defaults to
        application/json.
    :param registry: Registry of serializers, defaults to the global registry.
    :raises SerializerNotFound: If the serializer is not found for the content
        type.

    >>> serializer = Serializer.load()
    >>> serializer.encode('foo') == (b'"foo"', 'application/json', 'utf-8')
    True
    >>> serializer.decode(b'{"key": "value"}') == {'key': 'value'}
    True
    """

    default_content_type: str
    _serializers: dict[str, "SerializerPlugin"] = field(repr=False)

    @classmethod
    def load(cls, content_type="application/json", registry=None):
        """Load serializers from registry."""
        if registry is None or "serializers" not in registry:
            registry = registry_load("serializers", registry)

        serializers = registry.get("serializers", {})
        if content_type not in serializers:
            raise SerializerNotFound(f"Serializer {content_type!r} not found")

        return cls(content_type, serializers)

    def encode(self, data, content_type=None, optimize=False):
        """Encode data based on a content type.

        :param content_type: Content type to override the default value.
        :param optimize: Optimize the content type for string types.
        :return: The encoded payload, content type and charset.
        :raises SerializerNotFound: If the serializer is not found for the
            content type.
        """
        if optimize:
            if isinstance(data, bytes):
                content_type = "application/octet-stream"
            elif isinstance(data, str):
                content_type = "text/plain"

        if content_type is None:
            content_type = self.default_content_type

        try:
            serializer = self._serializers[content_type]
        except KeyError as error:
            raise SerializerNotFound(
                f"Serializer {content_type!r} not found"
            ) from error

        payload = serializer.encode(data)

        return payload, content_type, serializer.charset

    def decode(self, payload, content_type=None):
        """Decode a payload based on a content type.

        :param content_type: Content type to override the default value.
        :param charset: Character set of the payload.
        :return: The decoded data.
        :raises SerializerNotFound: If the serializer is not found for the
            content type or for the expected charset.
        """
        if content_type is None:
            content_type = self.default_content_type

        try:
            serializer = self._serializers[content_type]
        except KeyError as error:
            raise SerializerNotFound(
                f"Serializer {content_type!r} not found"
            ) from error

        return serializer.decode(payload)


@define(frozen=True)
class SerializerPlugin:
    """Serializer with an encode/decode method definable inline."""

    encode: Callable[[Any], bytes]
    decode: Callable[[bytes], Any]
    charset: str


json_serializer = SerializerPlugin(
    lambda data: json.dumps(
        data,
        sort_keys=True,
        indent=2,
        ensure_ascii=False,
    ).encode("utf-8"),
    lambda payload: json.loads(payload.decode("utf-8")),
    "utf-8",
)
"""Serializer for application/json."""


pickle_serializer = SerializerPlugin(
    pickle.dumps,
    pickle.loads,
    "binary",
)
"""Serializer for application/x-pickle."""


octet_stream_serializer = SerializerPlugin(
    lambda data: data,
    lambda payload: payload,
    "binary",
)
"""Serializer for application/octet-stream.

This serializer doesn't do anything.
"""


def _text_csv_serializer_encode(data, encoding="utf-8"):
    if not isinstance(data, list):
        data = [data]

    f = StringIO()
    keys = sorted(set().union(*data))
    writer = csv.DictWriter(f, keys, quoting=csv.QUOTE_NONNUMERIC)
    writer.writeheader()
    writer.writerows(data)
    return f.getvalue().encode(encoding)


def _text_csv_serializer_decode(payload, encoding="utf-8"):
    f = StringIO(payload.decode(encoding))
    reader = csv.DictReader(f, quoting=csv.QUOTE_NONNUMERIC)
    return list(reader)


text_csv_serializer = SerializerPlugin(
    _text_csv_serializer_encode,
    _text_csv_serializer_decode,
    "utf-8",
)
"""Serializer for text/csv."""


def _text_plain_serializer_encode(data, encoding="utf-8"):
    if not isinstance(data, list):
        data = [data]

    return "".join(f"{d}\n" for d in data).encode(encoding)


def _text_plain_serializer_decode(payload, encoding="utf-8"):
    return payload.decode("utf-8").rstrip("\n").split("\n")


text_plain_serializer = SerializerPlugin(
    _text_plain_serializer_encode,
    _text_plain_serializer_decode,
    "utf-8",
)
"""Serializer for text/plain."""

www_form_serializer = SerializerPlugin(
    lambda data: urlencode(data).encode("utf-8"),
    lambda payload: dict(parse_qsl(payload.decode("utf-8").strip())),
    "utf-8",
)
"""Serializer for application/x-www-form-urlencoded."""

yaml_serializer = SerializerPlugin(
    lambda data: yaml.safe_dump(
        data,
        sort_keys=True,
        allow_unicode=True,
    ).encode("utf-8"),
    lambda payload: yaml.safe_load(payload.decode("utf-8")),
    "utf-8",
)
"""Serializer for application/x-yaml."""



"""Unit tests for the data module."""

import pytest

from permaculture.data import (
    flatten,
    merge,
    unflatten,
)


@pytest.mark.parametrize(
    "nested, flat",
    [
        pytest.param(
            ["a", "b"],
            {"0": "a", "1": "b"},
            id="list",
        ),
        pytest.param(
            [["a", "b"], "c"],
            {"0/0": "a", "0/1": "b", "1": "c"},
            id="list > list",
        ),
        pytest.param(
            [{"a": 1, "b": 2}],
            {"0/a": 1, "0/b": 2},
            id="list > dict",
        ),
        pytest.param(
            {"a": 1},
            {"a": 1},
            id="dict",
        ),
        pytest.param(
            {"a": {"b": 1}},
            {"a/b": 1},
            id="dict > dict",
        ),
        pytest.param(
            {"a": ["b"]},
            {"a/0": "b"},
            id="dict > list",
        ),
        pytest.param(
            {"a": [{"b": 1}]},
            {"a/0/b": 1},
            id="dict > list > dict",
        ),
    ],
)
def test_flatten_unflatten(nested, flat):
    """Flattening and unflattening should be reversible."""
    assert unflatten(flatten(nested)) == nested
    assert flatten(unflatten(flat)) == flat


def test_unflatten_error():
    """Unflattening conflicting types should raise."""
    with pytest.raises(TypeError):
        unflatten({"a": "", "a/b": True})


@pytest.mark.parametrize(
    "dicts, merged",
    [
        pytest.param(
            [{"a": 1}, {"a": 1}],
            {"a": 1},
            id="same key, same value",
        ),
        pytest.param(
            [{"a": 1}, {"b": 1}],
            {"a": 1, "b": 1},
            id="different key, same value",
        ),
        pytest.param(
            [{"a": 1}, {"a": 2}],
            {"a": 1.5},
            id="same key, different value",
        ),
        pytest.param(
            [{"a": 1}, {"b": 2}],
            {"a": 1, "b": 2},
            id="different key, different value",
        ),
        pytest.param(
            [{"a": {"b": 1}}, {"a": {"b": 1}}],
            {"a": {"b": 1}},
            id="same key, same dict value",
        ),
        pytest.param(
            [{"a": {"b": 1}}, {"a": {"b": 2}}],
            {"a": {"b": 1.5}},
            id="same key, different dict value",
        ),
        pytest.param(
            [{"a": [1]}, {"a": [1]}],
            {"a": [1]},
            id="same key, same list value",
        ),
        pytest.param(
            [{"a": [1]}, {"a": [2]}],
            {"a": [1, 2]},
            id="same key, different list value",
        ),
        pytest.param(
            [{"a": "b"}, {"a": ""}],
            {"a": "b"},
            id="empty string value",
        ),
    ],
)
def test_merge(dicts, merged):
    """Merging dictionaries should return a merged dictionary."""
    result = merge(dicts)
    assert result == merged


@pytest.mark.parametrize(
    "dicts",
    [
        pytest.param(
            [{"a": 1}, {"a": "b"}],
            id="int str",
        ),
        pytest.param(
            [{"a": 1}, {"a": []}],
            id="int list",
        ),
        pytest.param(
            [{"a": "b"}, {"a": []}],
            id="str list",
        ),
    ],
)
def test_merge_error(dicts):
    """Merging data of different types should raise."""
    with pytest.raises(ValueError):
        merge(dicts)



"""Logger testing module."""

from contextlib import contextmanager
from datetime import UTC, datetime
from logging import Handler
from unittest.mock import patch


class LoggerHandler(Handler):
    """Logger handler for recording log messages."""

    def __init__(self):
        """Initialize an empty list of records."""
        super().__init__()
        self.records = []

    def emit(self, record):
        """Append to the list of records."""
        self.format(record)
        self.records.append(record)


@contextmanager
def logger_time(seconds=0.0):
    """Set the logger time to the given `seconds` within a context.

    :param seconds: Time as returned by `time.time()`.
    """
    with patch("logging.time.time") as mock_time, patch("logging.time.time_ns") as mock_time_ns:
        mock_time.return_value = seconds
        mock_time_ns.return_value = seconds
        yield datetime.fromtimestamp(seconds, UTC)



"""Integration tests for the http module."""

import pytest

from permaculture.http import HTTPCache, HTTPSession
from permaculture.storage import MemoryStorage


@pytest.fixture
def session():
    storage = MemoryStorage()
    return HTTPSession("http://httpbin.org").with_cache(storage, HTTPCache)


def test_respect_304(session):
    """The cache adapter should return the same response on 304."""
    r1 = session.get("/cache")
    r2 = session.get("/cache")

    assert r1 is r2


def test_respect_cache_control_max_age(session):
    """The cache adapter should return the same response within max-age."""
    r1 = session.get(
        "/response-headers",
        params={"Cache-Control": "max-age=3600"},
    )
    r2 = session.get(
        "/response-headers",
        params={"Cache-Control": "max-age=3600"},
    )

    assert r1 is r2


def test_respect_cache_control_no_cache(session):
    """The cache adapter should return the same response when no-cache."""
    r1 = session.get(
        "/response-headers",
        params={"Cache-Control": "no-cache"},
    )
    r2 = session.get(
        "/response-headers",
        params={"Cache-Control": "no-cache"},
    )

    assert r1 is not r2


def test_respect_expires(session):
    """The cache adapter should return the same response within expires."""
    r1 = session.get(
        "/response-headers",
        params={"Expires": "Sun, 01 Mar 2050 12:00:00 GMT"},
    )
    r2 = session.get(
        "/response-headers",
        params={"Expires": "Sun, 01 Mar 2050 12:00:00 GMT"},
    )

    assert r1 is r2



"""Plants For A Future database."""

import logging
from functools import partial
from pathlib import Path

import xlrd
from attrs import define, field

from permaculture.converter import Converter
from permaculture.database import DatabasePlant
from permaculture.locales import Locales
from permaculture.priority import LocationPriority, Priority

logger = logging.getLogger(__name__)


@define(frozen=True)
class PFAFFile:
    path: Path

    def main_database(self):
        wb = xlrd.open_workbook(self.path)
        return wb.sheet_by_name("MAIN DATABASE")


@define(frozen=True)
class PFAFConverter(Converter):
    locales: Locales = field(factory=partial(Locales.from_domain, "pfaf"))

    def convert_float(self, key, value):
        if isinstance(value, float | int):
            return [(self.translate(key), float(value))]
        else:
            return super().convert_float(key, value)

    def convert_item(self, key, value):
        dispatchers = {
            "Author": self.convert_ignore,
            "Common name": self.convert_list,
            "Cultivation details": self.convert_ignore,
            "Deciduous/Evergreen": self.convert_letters,
            "Drought": self.convert_ignore,
            "Edible uses": self.convert_ignore,
            "Growth rate": self.convert_list,
            "Habitat": self.convert_ignore,
            "Height": self.convert_float,
            "Known hazards": self.convert_ignore,
            "Latin name": self.convert_token,
            "Medicinal": self.convert_ignore,
            "Moisture": self.convert_letters,
            "Pollinators": self.convert_list,
            "Propagation": self.convert_ignore,
            "Range": self.convert_ignore,
            "Shade": self.convert_letters,
            "Soil": self.convert_letters,
            "Uses notes": self.convert_ignore,
            "Width": self.convert_float,
            "Wildlife": self.convert_ignore,
            "pH": self.convert_letters,
        }
        return dispatchers.get(key, self.convert_string)(key, value)


@define(frozen=True)
class PFAFModel:
    file: PFAFFile
    converter: PFAFConverter = field(factory=PFAFConverter)

    @classmethod
    def from_path(cls, path):
        file = PFAFFile(Path(path))
        return cls(file)

    def all_plants(self):
        try:
            ws = self.file.main_database()
        except FileNotFoundError as error:
            logger.info("Skipping Plants For A Future: %(error)s", {"error": error})
            return []

        rows = ws.get_rows()
        header = [h.value for h in next(rows)]
        for row in rows:
            yield self.converter.convert(
                {k: v.value for k, v in zip(header, row, strict=True)}
            )


@define(frozen=True)
class PFAFIngestor:
    model: PFAFModel
    priority: Priority = field(factory=Priority)

    @classmethod
    def from_config(cls, config):
        """Instantiate PFAFIngestor from config."""
        model = PFAFModel.from_path(config.pfaf_file)
        priority = LocationPriority("United Kingdom").with_cache(config.storage)
        return cls(model, priority)

    def fetch_all(self):
        count = 0
        for p in self.model.all_plants():
            count += 1
            if count % 100 == 0:
                logger.info("PFAF: ingested %d plants", count)
            yield DatabasePlant(p, self.priority.weight)
        logger.info("PFAF: ingested %d plants total", count)



"""Unique plugins."""

from http.cookiejar import Cookie, CookieJar

from permaculture.database import DatabasePlant
from permaculture.nlp import normalize


def unique_cookie(unique):
    """Generate a unique cookie."""
    version = unique("integer")
    name = unique("text")
    value = unique("text")
    port = None
    port_specified = False
    domain = unique("text")
    domain_specified = False
    domain_initial_dot = False
    path = unique("text")
    path_specified = True
    secure = True
    expires = unique("integer")
    discard = False
    comment = None
    comment_url = None
    rest = {"HttpOnly": None}
    rfc2109 = False
    return Cookie(
        version=version,
        name=name,
        value=value,
        port=port,
        port_specified=port_specified,
        domain=domain,
        domain_specified=domain_specified,
        domain_initial_dot=domain_initial_dot,
        path=path,
        path_specified=path_specified,
        secure=secure,
        expires=expires,
        discard=discard,
        comment=comment,
        comment_url=comment_url,
        rest=rest,
        rfc2109=rfc2109,
    )


def unique_cookies(unique, count=1):
    """Generate a cookie jar with unique cookies."""
    cookie_jar = CookieJar()
    for _ in range(count):
        cookie = unique("cookie")
        cookie_jar.set_cookie(cookie)

    return cookie_jar


def unique_plant(unique):
    """Generate a unique database plant."""
    return DatabasePlant(
        {
            "scientific name": unique("token"),
            "common name": unique("text"),
        }
    )


def unique_token(unique):
    """Generate a unique token."""
    return normalize(unique("text"))



"""Unit tests for the registry module."""
from importlib.metadata import EntryPoint
from unittest.mock import patch

import pytest

import permaculture.registry
from permaculture.registry import (
    registry_add,
    registry_get,
    registry_load,
    registry_remove,
)


def test_registry_load():
    """Entry points should be loaded into a registry dictionary."""
    with patch.object(
        permaculture.registry,
        "get_entry_points",
    ) as mock_entry_points:
        mock_entry_points.return_value = iter(
            [
                EntryPoint(
                    "registry", "permaculture.registry", "permaculture.group"
                ),
            ]
        )
        registry = registry_load("group")
        assert "registry" in registry["group"]
        assert registry["group"]["registry"] == permaculture.registry


def test_registry_load_ignore():
    """Entry points that fail to load should be ignored."""
    with patch.object(
        permaculture.registry,
        "get_entry_points",
    ) as mock_entry_points:
        mock_entry_points.return_value = iter(
            [
                EntryPoint(
                    "error", "permaculture.registry", "permaculture.group"
                ),
            ]
        )
        registry = registry_load("group")
        assert "test" not in registry


def test_registry_add():
    """Adding an entry in an empty registry should return a new registry."""
    registry = registry_add("group", "name", True)
    assert registry["group"]["name"]


def test_registry_add_existing():
    """Adding an entry in an existing registry should replace existing."""
    registry = {
        "group": {
            "name": False,
        },
    }
    registry = registry_add("group", "name", True, registry)
    assert registry["group"]["name"]


def test_registry_remove_existing():
    """Removing an existing entry from a registry should remove it."""
    registry = registry_add("group", "name1", True)
    registry = registry_add("group", "name2", True)
    registry_remove("group", "name1", registry)
    assert "name1" not in registry["group"]
    assert "name2" in registry["group"]


def test_registry_remove_non_existing():
    """Removing an existing entry from a registry should return silently."""
    registry = registry_add("group", "name2", True)
    registry_remove("group", "name1", registry)
    assert "name2" in registry["group"]


def test_registry_get():
    """Getting an existing entry from a registry should find it."""
    registry = registry_add("group", "name", True)
    assert registry_get("group", "name", registry)


def test_registry_get_error():
    """Getting a non-existing entry from a registry should raise."""
    with pytest.raises(KeyError):
        registry_get("group", "name", {})


@pytest.mark.parametrize(
    "group,name",
    [
        ("serializers", "application/json"),
        ("serializers", "text/plain"),
    ],
)
def test_registry_get_setup(group, name):
    """Getting from the registry should lookup entry points in setup.py."""
    registry_get(group, name)



"""Unit tests for the serializer module."""

from argparse import ArgumentParser

import pytest
from hamcrest import assert_that, has_properties

from permaculture.registry import registry_add
from permaculture.serializer import (
    Serializer,
    SerializerAction,
    SerializerNotFound,
    SerializerPlugin,
    octet_stream_serializer,
    text_csv_serializer,
    text_plain_serializer,
    www_form_serializer,
)

INT_TEST = 0
FLOAT_TEST = 3.1416
BYTES_TEST = b"bytes"
STRING_TEST = "abcdé\u8463"
LIST_TEST = [
    INT_TEST,
    FLOAT_TEST,
    STRING_TEST,
]
DICT_TEST = {
    "int": INT_TEST,
    "float": FLOAT_TEST,
    "string": STRING_TEST,
    "list": LIST_TEST,
}


@pytest.fixture(
    params=[
        "application/json",
        "application/x-pickle",
        "application/x-yaml",
    ]
)
def serializer(request):
    """Produce pytest parameters for all serializers."""
    return Serializer.load(request.param, {})


def test_serializer_action_default():
    """A SerializerAction should default to application/x-yaml."""
    parser = ArgumentParser()
    parser.add_argument("--serializer", action=SerializerAction)
    result = parser.parse_args([])

    assert_that(
        result,
        has_properties(
            serializer=has_properties(
                default_content_type="application/x-yaml"
            )
        ),
    )


def test_serializer_action_custom():
    """A SerializerAction should use the content-type given as argument."""
    parser = ArgumentParser()
    parser.add_argument("--serializer", action=SerializerAction)
    result = parser.parse_args(["--serializer", "text/plain"])

    assert_that(
        result,
        has_properties(
            serializer=has_properties(default_content_type="text/plain")
        ),
    )


def test_serializer_load_not_found():
    """Loading with an unknown content type should raise an exception."""
    content_type = "test"
    with pytest.raises(SerializerNotFound):
        Serializer.load(content_type)


def test_serializer_encode_optimized_bytes():
    """Encoding bytes with optimization should return an octet stream."""
    serializer = Serializer.load()
    _, content_type, charset = serializer.encode(BYTES_TEST, optimize=True)
    assert content_type == "application/octet-stream"
    assert charset == "binary"


def test_serializer_encode_optimized_string():
    """Encoding a string with optimization should return plain text."""
    serializer = Serializer.load()
    _, content_type, charset = serializer.encode(STRING_TEST, optimize=True)
    assert content_type == "text/plain"
    assert charset == "utf-8"


def test_serializer_encode_optimized_other():
    """Encoding other data with optimization should use the default."""
    content_type = "test"
    registry = registry_add(
        "serializers",
        content_type,
        octet_stream_serializer,
    )
    serializer = Serializer.load(content_type, registry)
    assert content_type == serializer.encode(LIST_TEST, optimize=True)[1]


def test_serializer_encode_content_type():
    """Encoding with a given content type should override the default."""
    registry = registry_add(
        "serializers",
        "application/octet-stream",
        octet_stream_serializer,
    )
    registry = registry_add(
        "serializers",
        "text/plain",
        text_plain_serializer,
        registry,
    )
    serializer = Serializer.load("text/plain", registry)
    content_type = serializer.encode(
        None, content_type="application/octet-stream"
    )[1]
    assert content_type == "application/octet-stream"
    data = serializer.decode(None, content_type="application/octet-stream")
    assert data is None


@pytest.mark.parametrize(
    "data",
    [
        pytest.param(
            INT_TEST,
            id="int",
        ),
        pytest.param(
            FLOAT_TEST,
            id="float",
        ),
        pytest.param(
            STRING_TEST,
            id="string",
        ),
        pytest.param(
            LIST_TEST,
            id="list",
        ),
        pytest.param(
            DICT_TEST,
            id="dict",
        ),
    ],
)
def test_serializer_encode_decode(data, serializer):
    """Encoding data and decoding the returned payload should be the same."""
    payload, _, _ = serializer.encode(data)
    assert isinstance(payload, bytes)
    assert serializer.decode(payload) == data


def test_serializer_encode_error(serializer):
    """A serializer should raise when encoding an unknown content type."""
    content_type = "test"
    with pytest.raises(SerializerNotFound):
        serializer.encode("", content_type)


def test_serializer_decode_error(serializer):
    """A serializer should raise when decoding an unknown content type."""
    content_type = "test"
    with pytest.raises(SerializerNotFound):
        serializer.decode(b"", content_type)


def test_serializer_exception():
    """A serializer that raises an exception should be re-raised."""

    class TestException(Exception):
        """Test exception."""

    def test_serializer(_):
        """Test serializer raising an exception."""
        raise TestException

    content_type = "test"
    registry = registry_add(
        "serializers",
        content_type,
        SerializerPlugin(test_serializer, test_serializer, "test"),
    )
    serializer = Serializer.load(content_type, registry)
    with pytest.raises(TestException):
        serializer.encode(None)
    with pytest.raises(TestException):
        serializer.decode(None)


def test_octet_stream_serializer():
    """The application/octet-stream content type should not encode a string."""
    payload = octet_stream_serializer.encode(BYTES_TEST)
    assert isinstance(payload, bytes)
    assert octet_stream_serializer.decode(payload) == BYTES_TEST


def test_text_csv_serializer():
    """The text/csv content type should encode a list of lists."""
    payload = text_csv_serializer.encode([{"a": 1}])
    assert isinstance(payload, bytes)
    assert text_csv_serializer.decode(payload) == [{"a": 1}]


def test_text_plain_serializer():
    """The text/plain content type should encode a string."""
    payload = text_plain_serializer.encode([STRING_TEST])
    assert isinstance(payload, bytes)
    assert text_plain_serializer.decode(payload) == [STRING_TEST]


@pytest.mark.parametrize(
    "data",
    [
        pytest.param(
            {},
            id="empty",
        ),
        pytest.param(
            {"key": "value"},
            id="single",
        ),
        pytest.param(
            {"key1": "value1", "key2": "value2"},
            id="multiple",
        ),
    ],
)
def test_www_form_serializer(data):
    """Should serialize and deserialize to the same data."""
    payload = www_form_serializer.encode(data)
    assert isinstance(payload, bytes)
    assert www_form_serializer.decode(payload) == data



"""Unit functions."""

from attrs import define, field


@define(frozen=True)
class Unit:
    multiplier: float = field
    offset: float = field(default=0.0)

    def __rmul__(self, other):
        return (other + self.offset) * self.multiplier


# Distance.
foot = feet = Unit(0.3048)
inch = inches = Unit(0.0254)
km = kilometer = kilometers = Unit(1000.0)
meter = meters = Unit(1.0)
mile = miles = Unit(1609.34)

# Temperature.
celsius = Unit(1.0)
fahrenheit = Unit(5 / 9, -32)



"""Natural language processing functions."""

import re
import string
from argparse import ArgumentTypeError

from attrs import define, field
from more_itertools import windowed
from unidecode import unidecode


def normalize(text):
    """Transform text to a more consistent and simplified format."""
    # Strip leading and trailing spaces.
    text = text.strip()

    # Remove parentheses.
    text = re.sub(r"\([^)]*\)", "", text)

    # Remove brackets.
    text = re.sub(r"\[[^]]*\]", "", text)

    # Remove unicode characters.
    text = unidecode(text)

    # Lower case.
    text = text.lower()

    # Remove punctuation.
    text = " ".join(re.findall(r"\w+", text))

    # Remove words smaller than a minimum length.
    text = " ".join(w for w in text.split() if len(w) > 1)

    # Remove digits.
    text = " ".join(w for w in text.split() if w not in string.digits)

    return text


def lev(s1, s2, sort=True):
    """Calculates the normalized Levenshtein distance between two strings.

    https://en.wikipedia.org/wiki/Levenshtein_distance
    """
    if sort:
        s1 = " ".join(sorted(s1.split()))
        s2 = " ".join(sorted(s2.split()))

    l1, l2 = len(s1), len(s2)
    matrix = [*range(l1 + 1)] * (l2 + 1)
    for zz in range(l2 + 1):
        matrix[zz] = [*range(zz, zz + l1 + 1)]

    for zz in range(l2):
        for sz in range(l1):
            if s1[sz] == s2[zz]:
                matrix[zz + 1][sz + 1] = min(
                    matrix[zz + 1][sz] + 1,
                    matrix[zz][sz + 1] + 1,
                    matrix[zz][sz],
                )
            else:
                matrix[zz + 1][sz + 1] = min(
                    matrix[zz + 1][sz] + 1,
                    matrix[zz][sz + 1] + 1,
                    matrix[zz][sz] + 1,
                )

    distance = float(matrix[l2][l1])
    return 1.0 - distance / max(l1, l2)


def score(s1, s2):
    return max(
        lev(s1, " ".join(filter(None, w2)))
        for w2 in windowed(s2.split(), len(s1.split()))
    )


def score_type(x):
    """Argument parser score type."""
    try:
        x = float(x)
    except ValueError as e:
        raise ArgumentTypeError(f"{x} not a floating-point literal") from e

    if x < 0.0 or x > 1.0:
        raise ArgumentTypeError(f"{x} not in range [0.0, 1.0]")

    return x


@define(frozen=True)
class Extractor:
    query = field()
    normalizer = field(default=lambda x: x)
    scorer = field(default=lambda *_: 1)

    @property
    def normalized_query(self):
        return self.normalizer(self.query)

    def extract(self, choices):
        """Extract the score for each choice."""
        normalized_query = self.normalized_query
        for choice in choices:
            normalized_choice = self.normalizer(choice)
            score = self.scorer(normalized_query, normalized_choice)
            yield (score, choice)

    def extract_one(self, choices):
        """Extract the beschoice with the best score.

        :raises ValueError: When there are no choices.
        """
        results = self.extract(choices)
        return max(results, key=lambda i: i[0])

    def choose(self, choices, default=None):
        result = self.extract_one(choices)
        return result[1] if result else default



"""Storage providers."""

import logging
import sqlite3
from collections.abc import MutableMapping
from hashlib import md5
from pathlib import Path
from urllib.parse import quote, unquote

from attrs import define, field

from permaculture.action import SingleAction
from permaculture.serializer import Serializer, json_serializer

logger = logging.getLogger(__name__)


def hash_request(method, url, body=None):
    """Hash an HTTP-like request into a storage key."""
    data = json_serializer.encode({"method": method, "url": url, "body": body})
    return md5(data).hexdigest()  # noqa: S324


class StorageAction(SingleAction):
    """Argument action for storage."""

    metavar = "PATH"

    def __init__(self, option_strings, registry=None, **kwargs):
        """Initializer storage defaults."""
        default = kwargs.pop("default", None)
        kwargs.setdefault("default", self.get_storage(default))
        kwargs.setdefault("metavar", self.metavar)
        kwargs.setdefault("help", f"storage path (default {default})")
        super().__init__(option_strings, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        """Set the values to a storage."""
        storage = self.get_storage(values)

        super().__call__(parser, namespace, storage, option_string)

    @classmethod
    def get_storage(cls, path):
        """Get storage with a default path."""
        return Storage.load(path)


class Storage(MutableMapping):
    @classmethod
    def load(cls, path=None):
        return FileStorage(path) if path else MemoryStorage()


MemoryStorage: Storage = dict


@define(frozen=True)
class _NullStorage(Storage):
    """Null storage.

    This storage stores a value and always retrieves the default value.
    """

    def __getitem__(self, key):
        """Raise KeyError."""
        raise KeyError(key)

    def __setitem__(self, key, value):
        """Noop."""

    def __delitem__(self, key):
        """Raise KeyError."""
        raise KeyError(key)

    def __iter__(self):
        yield from ()

    def __len__(self):
        """Return 0."""
        return 0


null_storage = _NullStorage()


@define(frozen=True)
class FileStorage(Storage):
    """File storage.

    :param base_dir: Base directory for storing files.
    :param serializer: Serializer, defaults to `application/x-pickle`.
    """

    base_dir: Path = field(converter=Path)
    serializer: Serializer = field(
        default="application/x-pickle",
        converter=lambda x: (
            x if isinstance(x, Serializer) else Serializer.load(x)
        ),
    )

    def key_to_path(self, key):
        path = self.base_dir / quote(key, "")
        return path

    def path_to_key(self, path):
        return unquote(path.name)

    def __getitem__(self, key):
        """Read from file."""
        path = self.key_to_path(key)
        if not path.exists():
            raise KeyError(key)

        logger.debug("reading from %(path)s", {"path": path})
        payload = path.read_bytes()
        return self.serializer.decode(payload)

    def __setitem__(self, key, value):
        """Write to file."""
        path = self.key_to_path(key)
        payload, *_ = self.serializer.encode(value)
        path.parent.mkdir(parents=True, exist_ok=True)
        logger.debug("writing to %(path)s", {"path": path})
        path.write_bytes(payload)

    def __delitem__(self, key):
        """Unlink file."""
        try:
            self.key_to_path(key).unlink()
        except FileNotFoundError as error:
            raise KeyError(key) from error

    def __iter__(self):
        return map(self.path_to_key, self.base_dir.iterdir())

    def __len__(self):
        return sum(1 for _ in self)


@define(frozen=True)
class SqliteStorage(Storage):
    """Sqlite storage.

    :param path: Path to SQLite database.
    :param serializer: Serializer, defaults to `application/x-pickle`.
    """

    conn = field()
    serializer: Serializer = field(
        default="application/x-pickle",
        converter=lambda x: (
            x if isinstance(x, Serializer) else Serializer.load(x)
        ),
    )

    @classmethod
    def from_path(cls, path):
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)

        conn = sqlite3.connect(path)
        conn.execute("CREATE TABLE IF NOT EXISTS storage (key, data)")
        conn.execute(
            "CREATE UNIQUE INDEX IF NOT EXISTS idx_storage_key "
            "ON storage (key)"
        )
        conn.commit()

        return cls(conn)

    def __getitem__(self, key):
        cursor = self.conn.execute(
            "SELECT data FROM storage WHERE key = ?", (key,)
        )
        row = cursor.fetchone()
        if row is None:
            raise KeyError(key)

        return self.serializer.decode(row[0])

    def __setitem__(self, key, value):
        data, *_ = self.serializer.encode(value)
        self.conn.execute("INSERT INTO storage VALUES (?, ?)", (key, data))
        self.conn.commit()

    def __delitem__(self, key):
        cursor = self.conn.execute(
            "DELETE FROM storage WHERE key = ? RETURNING 1", (key,)
        )
        row = cursor.fetchone()
        if row is None:
            raise KeyError(key)

        self.conn.commit()

    def __iter__(self):
        cursor = self.conn.execute("SELECT key FROM storage")
        for row in cursor.fetchmany():
            yield row[0]

    def __len__(self):
        cursor = self.conn.execute("SELECT COUNT(key) FROM storage")
        row = cursor.fetchone()
        return row[0]



"""Unit tests for the ipinfo module."""

from unittest.mock import Mock

import pytest

from permaculture.ipinfo import IPinfo
from permaculture.location import LocationPoint

from .stubs import StubRequestsResponse


def test_ipinfo_with_cache():
    """Instantiating an IPinfo instance with cache should use the cache."""
    storage = {}
    session = Mock()
    IPinfo(session).with_cache(storage)
    session.with_cache.assert_called_once_with(storage)


@pytest.mark.parametrize(
    "ip, path",
    [
        (None, "/json"),
        ("1.2.3.4", "/1.2.3.4/json"),
    ],
)
def test_ipinfo_json(ip, path):
    """The IP json should GET with the given parameters."""
    json = Mock(return_value={})
    session = Mock(get=Mock(return_value=StubRequestsResponse(json=json)))
    IPinfo(session).json(ip)
    session.get.assert_called_once_with(path)


def test_ipinfo_point():
    """A point should get the loc."""
    json = Mock(return_value={"loc": "1.0,2.0"})
    session = Mock(get=Mock(return_value=StubRequestsResponse(json=json)))
    point = IPinfo(session).point(None)
    assert point == LocationPoint(1.0, 2.0)



"""Natural Capital website."""

import logging
import re
from functools import partial
from pathlib import Path

from attrs import define, field
from bs4 import BeautifulSoup
from requests import Session
from yarl import URL

from permaculture.converter import FLOAT_RE, Converter
from permaculture.database import DatabasePlant
from permaculture.http import HTTPCacheAdapter, HTTPCacheAll, HTTPSession
from permaculture.locales import Locales
from permaculture.priority import LocationPriority, Priority
from permaculture.storage import null_storage
from permaculture.unit import feet, inches

NC_ORIGIN = "https://permacultureplantdata.com"

logger = logging.getLogger(__name__)


class NCAuthenticationError(Exception):
    """Raised when authentication fails."""


@define(frozen=True)
class NCAuthentication:
    username: str = field()
    password: str = field()
    session: Session = field(factory=Session)

    def get_payload(self, request):
        response = self.session.get(request.url)
        bs = BeautifulSoup(response.text, "html.parser")
        inputs = bs.find("form", attrs={"name": "btl-formlogin"})
        return {
            "bttask": inputs.find(attrs={"name": "bttask"}).get("value"),
            "username": self.username,
            "passwd": self.password,
            "return": inputs.find(attrs={"name": "return"}).get("value"),
            inputs.find(attrs={"value": "1"}).get("name"): "1",
            "remember": "yes",
        }

    def authenticate(self, request):
        payload = self.get_payload(request)
        response = self.session.post(request.url, data=payload)
        if not response.cookies:
            raise NCAuthenticationError(
                f"Failed authenticating to {request.url}"
            )

        request.prepare_cookies(response.cookies)
        request.headers["X-Cache-All"] = "overwrite"
        return response.headers["Set-Cookie"]


class NCAdapter(HTTPCacheAdapter):
    def __init__(self, authentication, **kwargs):
        super().__init__(**kwargs)
        self.authentication = authentication

    def send(self, request, *args, **kwargs):  # pragma: no cover
        response = super().send(request, *args, **kwargs)
        if response.headers.get("X-Logged-In", "False") == "False":
            cookie = self.authentication.authenticate(request)
            response = super().send(request, *args, **kwargs)
            response.raw._original_response.msg.add_header(
                "Set-Cookie", cookie
            )

        return response


@define(frozen=True)
class NCWeb:
    session: HTTPSession = field(factory=partial(HTTPSession, NC_ORIGIN))

    def view_detail(self, Id):
        """View plant detail."""
        response = self.session.post(
            "/plant-database/new-the-plant-list",
            params={
                "vw": "detail",
                "id": Id,
            },
        )
        return response.text

    def view_list(self, sci_name="", sort_name="", limit_start=0):
        """View plant list."""
        response = self.session.post(
            "/plant-database/new-the-plant-list",
            params={
                "vw": "list",
            },
            data={
                "sortName": sort_name,
                "sciName": sci_name,
                "bfilter": "Set Filter",
                "limitstart": limit_start,
            },
        )
        return response.text


@define(frozen=True)
class NCConverter(Converter):
    locales: Locales = field(factory=partial(Locales.from_domain, "nc"))

    def convert_period(self, key, value):
        if value is None or not value.strip():
            return []

        k = self.translate(key)
        n = [self.translate(v, key) for v in re.split(r"\s*-\s*", value)]
        match len(n):
            case 1:
                return [(f"{k}/min", n[0]), (f"{k}/max", n[0])]
            case 2:
                return [(f"{k}/min", n[0]), (f"{k}/max", n[1])]
            case _:
                raise ValueError(f"Unknown period: {value!r}")

    def convert_range(self, key, value, unit=1.0):
        if m := re.match(rf"{FLOAT_RE} (?P<u>\w+)", value):
            u = m.group("u")
            match u:
                case "inches":
                    unit = inches
                case "feet":
                    unit = feet
                case _:
                    raise ValueError(f"Unknown range unit: {u!r}")

        return super().convert_range(key, value, unit)

    def convert_string(self, key, value):
        if value == "(none)":
            return [(key, None)]
        else:
            return super().convert_string(key, value)

    def convert_item(self, key, value):
        dispatchers = {
            "Bacteria-Fungal Ratio": self.convert_ignore,
            "Bloom Time": self.convert_period,
            "Common name": self.convert_list,
            "Compatible": self.convert_bool,
            "Fire Damage": self.convert_list,
            "Flood": self.convert_list,
            "Flower Color": self.convert_list,
            "Fruit Time": self.convert_period,
            "Growth Rate": self.convert_list,
            "Height": self.convert_range,
            "Minimum Root Depth": partial(self.convert_float, unit=inches),
            "Notes": self.convert_ignore,
            "Root Type": self.convert_list,
            "Soil Moisture": self.convert_list,
            "Soil Type": self.convert_list,
            "Soil pH": self.convert_range,
            "Spread": self.convert_range,
            "Sun": self.convert_list,
            "USDA Hardiness Zones": self.convert_range,
        }
        return dispatchers.get(key, self.convert_string)(key, value)


@define(frozen=True)
class NCModel:
    web: NCWeb = field(factory=NCWeb)
    converter: NCConverter = field(factory=NCConverter)

    def with_authentication(
        self, username: str, password: str, storage=null_storage
    ):
        """Add authentication credentials to the model."""
        cache = HTTPCacheAll(storage)
        authentication = NCAuthentication(username, password)
        adapter = NCAdapter(authentication, cache=cache)
        self.web.session.with_adapter(adapter)
        return self

    def parse_tr(self, row):
        """Parse a table row into an iterator of text or links."""
        for td in row.find_all("td"):
            text = td.get_text()
            a = td.find("a")
            if a:
                yield NCLink(text, URL(a["href"]))
            else:
                yield text

    def parse_table(self, table):
        """Parse a table with a header into a list of dictionaries."""
        header, *trs = table.find_all("tr")
        keys = [td.get_text().strip() for td in header.find_all("td")]
        for tr in trs:
            values = self.parse_tr(tr)
            yield dict(zip(keys, values, strict=True))

    def parse_plant_list(self, text):
        """Parse a plant list consisting of an HTML table of plants."""
        soup = BeautifulSoup(text, "html.parser")
        tables = soup.select("table[width='100%']")
        for table in tables:
            if table.find("td", attrs={"class": "plantList"}):
                yield from self.parse_table(table)

    def parse_items(self, table):
        """Parse the items total inside a bold tag."""
        b = table.find("b")
        m = re.match(r"Items \d+ - \d+ of (?P<total>\d+)", b.text)
        if not m:
            raise ValueError(f"Expected Items, got {b.text!r}")

        return int(m.group("total"))

    def parse_plant_total(self, text):
        """Parse a plant list consisting of an HTML table of plants."""
        soup = BeautifulSoup(text, "html.parser")
        tables = soup.select("table[width='100%']")
        for table in tables:
            if table.find("td", attrs={"width": "50%"}):
                return self.parse_items(table)

        return -1

    def get_plant_total(self):
        """Get the total number of plants."""
        text = self.web.view_list()
        return self.parse_plant_total(text)

    def parse_key_value(self, key_value):
        """Parse a bold key and roman value."""
        prefix = key_value.find("b").text
        key = prefix.strip(": ")
        value = key_value.text.removeprefix(prefix).strip()
        return key, value

    def parse_detail(self, text):
        """Parse a plant detail consisting of an HTML table and paragraphs."""
        soup = BeautifulSoup(text, "html.parser")
        tables = soup.select("table[width='100%']:not([class*=list])")
        detail = {}
        for table in tables:
            for tr in table.find_all("tr"):
                for td in tr.find_all("td"):
                    ps = td.find_all("p")
                    if ps:
                        for p in ps:
                            key, value = self.parse_key_value(p)
                            detail[key] = value

                        continue

                    b = td.find("b")
                    if b:
                        key, value = self.parse_key_value(td)
                        detail[key] = value
                        continue

        return detail

    def get_plant(self, Id):
        """Get plant details by Id."""
        detail = self.web.view_detail(Id)
        data = self.parse_detail(detail)
        return self.converter.convert(data)

    def get_plants(self, sci_name="", sort_name="", limit_start=0):
        """Get plants by scientific name or by common name."""
        try:
            plant_list = self.web.view_list(sci_name, sort_name, limit_start)
        except NCAuthenticationError as error:
            logger.info(
                "Skipping Natural Capital: %(error)s", {"error": error}
            )
            return

        for table in self.parse_plant_list(plant_list):
            yield self.converter.convert(table)


@define(frozen=True)
class NCLink:
    text: str
    url: URL = field(converter=URL)

    @property
    def Id(self):
        return int(self.url.query["id"])

    def __str__(self):
        return self.text


@define(frozen=True)
class NCIngestor:
    model: NCModel = field(factory=NCModel)
    priority: Priority = field(factory=Priority)

    @classmethod
    def from_config(cls, config):
        """Instantiate NCIngestor from config."""
        password = config.nc_password
        if config.nc_password_file:
            password = Path(config.nc_password_file).read_text().strip()
        model = NCModel().with_authentication(
            config.nc_username,
            password,
            config.storage,
        )
        priority = LocationPriority("United States").with_cache(config.storage)
        return cls(model, priority)

    def fetch_all(self):
        total = self.model.get_plant_total()
        logger.info("NC: found %d plants total", total)
        count = 0
        for limit_start in range(0, total, 50):
            for plant in self.model.get_plants(limit_start=limit_start):
                count += 1
                if count % 100 == 0:
                    logger.info("NC: ingested %d/%d plants", count, total)
                yield DatabasePlant(
                    self.model.get_plant(plant["plant name"].Id),
                    self.priority.weight,
                )
        logger.info("NC: ingested %d plants total", count)



"""Database utilities."""

import json
import sqlite3
from collections.abc import Iterator, Mapping
from itertools import groupby, starmap
from operator import attrgetter, mul
from pathlib import Path

from attrs import define, field

from permaculture.data import merge
from permaculture.nlp import Extractor, normalize, score
from permaculture.storage import FileStorage


@define(frozen=True, hash=False)
class DatabasePlant(Mapping):
    """Plant record with structured data and weight."""

    data: dict = field(factory=dict)
    weight: float = 1.0

    @property
    def scientific_name(self):
        return self.data.get("scientific name", "")

    @property
    def common_names(self):
        return [
            key.removeprefix("common name/")
            for key in self.data
            if key.startswith("common name/")
        ]

    @property
    def names(self):
        return [self.scientific_name, *self.common_names]

    def with_database(self, name):
        """Add the database name to this plant."""
        self.data[f"database/{name}"] = True
        return self

    def __getitem__(self, key):
        return self.data[key]

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)


@define(frozen=True)
class Database:
    """Local SQLite database for plant records."""

    db_path: Path = field(converter=Path)

    @classmethod
    def from_storage(cls, storage):
        """Create a Database from a storage provider."""
        if isinstance(storage, FileStorage):
            return cls(storage.base_dir / "permaculture.db")
        return cls(":memory:")

    def _connect(self):
        self.db_path.parent.mkdir(parents=True, exist_ok=True)
        return sqlite3.connect(self.db_path)

    def _extract(self, query, choices):
        return Extractor(query, normalize, score).extract_one(choices)[0]

    def initialize(self):
        """Create tables and indexes."""
        with self._connect() as conn:
            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS plants (
                    id INTEGER PRIMARY KEY,
                    source TEXT NOT NULL,
                    scientific_name TEXT NOT NULL,
                    data TEXT NOT NULL,
                    weight REAL NOT NULL DEFAULT 1.0,
                    updated_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
                )
            """
            )
            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS common_names (
                    plant_id INTEGER REFERENCES plants(id),
                    name TEXT NOT NULL
                )
            """
            )
            conn.execute(
                "CREATE INDEX IF NOT EXISTS idx_sci" " ON plants(scientific_name)"
            )
            conn.execute("CREATE INDEX IF NOT EXISTS idx_cn" " ON common_names(name)")
            conn.execute(
                "CREATE VIRTUAL TABLE IF NOT EXISTS names_fts"
                " USING fts5(name, plant_id UNINDEXED, tokenize='trigram')"
            )

    def write_batch(self, source, records):
        """Persist a batch of plant records."""
        with self._connect() as conn:
            for record in records:
                data_json = json.dumps(record.data)
                cur = conn.execute(
                    "INSERT INTO plants"
                    " (source, scientific_name, data, weight)"
                    " VALUES (?, ?, ?, ?)",
                    (source, record.scientific_name, data_json, record.weight),
                )
                pid = cur.lastrowid
                common_names = record.common_names
                conn.executemany(
                    "INSERT INTO common_names (plant_id, name)" " VALUES (?, ?)",
                    [(pid, n) for n in common_names],
                )
                fts_rows = [
                    (record.scientific_name, pid),
                    *((n, pid) for n in common_names),
                ]
                conn.executemany(
                    "INSERT INTO names_fts (name, plant_id)" " VALUES (?, ?)",
                    fts_rows,
                )

    def sources(self, include=None) -> list[str]:
        """Return the distinct source names in the database.

        :param include: Optional regex to filter sources.
        """
        with self._connect() as conn:
            all_sources = [
                row[0] for row in conn.execute("SELECT DISTINCT source FROM plants")
            ]

        if include is None:
            return all_sources

        return [s for s in all_sources if include.match(s)]

    def iterate(self) -> Iterator[DatabasePlant]:
        """Iterate over all plants, merging across sources."""
        return _merge_all(
            plant.with_database(source) for source, plant in self._iterate_raw()
        )

    def _iterate_raw(self):
        """Yield (source, plant) pairs for all rows."""
        with self._connect() as conn:
            for source, data, weight in conn.execute(
                "SELECT source, data, weight FROM plants"
            ):
                yield source, DatabasePlant(json.loads(data), weight)

    def lookup(self, names: list[str], score: float) -> Iterator[DatabasePlant]:
        """Lookup characteristics by scientific names, merging across sources."""
        if not names:
            return

        with self._connect() as conn:
            placeholders = ",".join("?" * len(names))
            rows = conn.execute(
                "SELECT source, data, weight FROM plants"  # noqa: S608
                f" WHERE scientific_name IN ({placeholders})",
                names,
            )
            yield from _merge_all(
                DatabasePlant(json.loads(data), weight).with_database(src)
                for src, data, weight in rows
            )

    def search(self, name: str, score: float) -> Iterator[DatabasePlant]:
        """Search for plants by name using FTS5 trigram matching."""
        with self._connect() as conn:
            rows = conn.execute(
                "SELECT DISTINCT p.source, p.data, p.weight"
                " FROM names_fts fts"
                " JOIN plants p ON fts.plant_id = p.id"
                " WHERE fts.name MATCH ?",
                (f'"{name}"',),
            )
            yield from _merge_all(
                plant.with_database(src)
                for src, data, weight in rows
                if self._extract(
                    name,
                    (plant := DatabasePlant(json.loads(data), weight)).names,
                )
                >= score
            )


def _merge_all(
    plants: Iterator[DatabasePlant],
) -> Iterator[DatabasePlant]:
    """Group plants by scientific name, merging numbers and strings."""
    keyfunc = attrgetter("scientific_name")
    return (_merge(p) for _, p in groupby(sorted(plants, key=keyfunc), keyfunc))


def _merge(plants: Iterator[DatabasePlant]) -> DatabasePlant:
    """Merge plants with the same scientific name using weighted resolution."""
    plants = list(plants)

    def resolve(key, values):
        weights = [p.weight for p in plants if key in p]
        if isinstance(values[0], float | int):
            value = sum(starmap(mul, zip(weights, values, strict=True))) / sum(weights)
        else:
            _, value = max(zip(weights, values, strict=True))

        return value

    weights = [p.weight for p in plants]
    avg_weight = sum(weights) / len(weights) if weights else 1.0
    return DatabasePlant(merge(plants, resolve), avg_weight)



"""Unit tests for the converter module."""

import pytest

from permaculture.converter import Converter
from permaculture.locales import Locales


@pytest.fixture
def converter():
    """Return a testing Converter."""
    locales = Locales.from_domain("testing")
    return Converter(locales)


@pytest.mark.parametrize(
    "value, expected",
    [
        ("Yes", True),
        ("Y", True),
        ("No", False),
        ("N", False),
    ],
)
def test_converter_convert_bool(converter, value, expected):
    """Converting a boolean should parse Yes and No."""
    result = converter.convert_bool("", value)
    assert result == [("", expected)]


def test_converter_convert_bool_error(converter):
    """Converting an unknown boolean should raise."""
    with pytest.raises(ValueError):
        converter.convert_bool("", "test")


@pytest.mark.parametrize(
    "item, expected",
    [
        pytest.param(
            ("key", "1"),
            [("key", 1.0)],
            id="1",
        ),
        pytest.param(
            ("key", "+1"),
            [("key", 1.0)],
            id="+1",
        ),
        pytest.param(
            ("key", "-1"),
            [("key", -1.0)],
            id="-1",
        ),
        pytest.param(
            ("key", "1."),
            [("key", 1.0)],
            id="1.",
        ),
        pytest.param(
            ("key", "1.0"),
            [("key", 1.0)],
            id="1.0",
        ),
        pytest.param(
            ("key", ""),
            [],
            id="<empty>",
        ),
    ],
)
def test_converter_convert_float(converter, item, expected):
    """Converting a float should parse strings."""
    result = converter.convert_float(*item)
    assert result == expected


def test_converter_convert_float_unit(converter):
    """Converting a float should parse strings."""
    result = converter.convert_float("", "5.0", 2.0)
    assert result == [("", 10.0)]


def test_converter_convert_float_error(converter):
    """Converting an unknown float should raise."""
    with pytest.raises(ValueError):
        converter.convert_float("", "test")


def test_converter_convert_ignore(converter):
    """Converting an ignore item should return an empty list."""
    result = converter.convert_ignore("key", "value")
    assert result == []


@pytest.mark.parametrize(
    "value, expected",
    [
        ("1", 1),
        (1, 1),
    ],
)
def test_converter_convert_int(converter, value, expected):
    """Converting a int should parse strings and ints."""
    result = converter.convert_int("", value)
    assert result == [("", expected)]


@pytest.mark.parametrize(
    "item, expected",
    [
        pytest.param(
            ("key", "A"),
            [("key/a", True)],
            id="A",
        ),
        pytest.param(
            ("key", "AB"),
            [("key/a", True), ("key/b", True)],
            id="AB",
        ),
        pytest.param(
            ("key", "Ab"),
            [("key/ab", True)],
            id="Ab",
        ),
        pytest.param(
            ("key", "AbC"),
            [("key/ab", True), ("key/c", True)],
            id="AbC",
        ),
    ],
)
def test_converter_convert_letters(converter, item, expected):
    """Converting letters should split into each letter."""
    result = converter.convert_letters(*item)
    assert result == expected


@pytest.mark.parametrize(
    "item, expected",
    [
        pytest.param(
            ("key", "A"),
            [("key/a", True)],
            id="A",
        ),
        pytest.param(
            ("key", "A,B"),
            [("key/a", True), ("key/b", True)],
            id="A,B",
        ),
        pytest.param(
            ("key", "A, B"),
            [("key/a", True), ("key/b", True)],
            id="A, B",
        ),
    ],
)
def test_converter_convert_list(converter, item, expected):
    """Converting list should split into each letter."""
    result = converter.convert_list(*item)
    assert result == expected


@pytest.mark.parametrize(
    "item, expected",
    [
        pytest.param(
            ("key", "1 inches - 2 inches"),
            [
                ("key/min", 1),
                ("key/max", 2),
            ],
            id="unit",
        ),
        pytest.param(
            ("key", "1.0 - 2.0"),
            [
                ("key/min", 1.0),
                ("key/max", 2.0),
            ],
            id="dash",
        ),
        pytest.param(
            ("key", "1.0 \u2013 2.0"),
            [
                ("key/min", 1.0),
                ("key/max", 2.0),
            ],
            id="endash",
        ),
        pytest.param(
            ("key", "1.0"),
            [
                ("key/min", 1.0),
                ("key/max", 1.0),
            ],
            id="single value",
        ),
    ],
)
def test_converter_convert_range(converter, item, expected):
    """Converting a range should support single and double values."""
    result = converter.convert_range(*item)
    assert result == expected


@pytest.mark.parametrize(
    "item, expected",
    [
        pytest.param(
            ("key", "string"),
            [("key", "string")],
            id="string",
        ),
        pytest.param(
            ("key", 1),
            [("key", 1)],
            id="int",
        ),
    ],
)
def test_converter_convert_string(converter, item, expected):
    """Converting a string should not convert other types."""
    result = converter.convert_string(*item)
    assert result == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        (
            "key a.",
            "key",
        ),
        ("key (paren)", "key"),
    ],
)
def test_converter_convert_token(converter, value, expected):
    """Converting a token should normalize the value."""
    result = converter.convert_token("", value)
    assert result == [("", expected)]



"""La Société des Plantes database."""

import logging
import re
from functools import partial

from attrs import define, evolve, field
from bs4 import BeautifulSoup
from defusedxml.ElementTree import fromstring

from permaculture.converter import Converter
from permaculture.database import DatabasePlant
from permaculture.http import HTTPSession
from permaculture.locales import Locales
from permaculture.nlp import normalize
from permaculture.priority import LocationPriority, Priority

SDP_ORIGIN = "https://www.lasocietedesplantes.com"

SITEMAP_NS = "http://www.sitemaps.org/schemas/sitemap/0.9"

NON_PLANT_SLUGS = re.compile(
    r"kit-|certificat-|livre-|affiche-|t-shirt-|sweatshirt-"
    r"|atelier-|zine-|hydrolat|visite|rituel-|poivre-d"
    r"|camomille-allemande-sechee|basilic-sacre-rama-seche"
    r"|feuilles-de-framboisier-sechees|melisse-sechee"
    r"|sweatshirt-"
)

logger = logging.getLogger(__name__)


@define(frozen=True)
class SDPWeb:
    """La Société des Plantes web interface."""

    session: HTTPSession = field(factory=partial(HTTPSession, SDP_ORIGIN))

    def with_cache(self, storage):
        session = self.session.with_cache(storage)
        return evolve(self, session=session)

    def product_sitemap(self):
        """Fetch the product sitemap XML."""
        response = self.session.get("/product-sitemap.xml")
        return response.text

    def product_page(self, path):
        """Fetch an individual product page."""
        response = self.session.get(path)
        return response.text


def parse_sitemap_urls(xml_text):
    """Extract French product paths from the sitemap XML.

    >>> urls = parse_sitemap_urls(
    ...     '<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">'
    ...     '<url><loc>https://www.lasocietedesplantes.com/produits/a/</loc></url>'
    ...     '</urlset>'
    ... )
    >>> urls
    ['/produits/a/']
    """
    root = fromstring(xml_text)
    urls = []
    for url_el in root.findall(f"{{{SITEMAP_NS}}}url"):
        loc = url_el.findtext(f"{{{SITEMAP_NS}}}loc", "")
        if "/produits/" in loc:
            path = loc.removeprefix(SDP_ORIGIN)
            slug = path.removeprefix("/produits/").rstrip("/")
            if not NON_PLANT_SLUGS.search(slug):
                urls.append(path)
    return urls


def parse_scientific_name(text):
    r"""Extract scientific name from a product description line.

    The first line typically looks like: "Withania somnifera, Solanacées"

    >>> parse_scientific_name("Withania somnifera, Solanacées")
    'withania somnifera'
    >>> parse_scientific_name("Borago officinalis alba, Boraginacées")
    'borago officinalis alba'
    >>> parse_scientific_name("No match here")
    """
    m = re.match(
        r"([A-Z][a-z]+ [a-z]+(?:\s+(?:var\.|subsp\.)\s+[a-z]+|\s+[a-z]+)*)"
        r"\s*,",
        text.strip(),
    )
    if m:
        return normalize(m.group(1))
    return None


def parse_family(text):
    r"""Extract the plant family from a description line.

    >>> parse_family("Withania somnifera, Solanacées")
    'solanacées'
    >>> parse_family("No match here")
    """
    m = re.match(
        r"[A-Z][a-z]+ [a-z].*,\s*([A-ZÀ-Ý][a-zà-ÿ]+(?:cées|acées))",
        text.strip(),
    )
    if m:
        return m.group(1).lower()
    return None


def parse_product(html):
    """Parse a product page and return a dict of raw fields.

    Returns None for pages that don't look like plant products.
    """
    soup = BeautifulSoup(html, "html.parser")

    title_el = soup.find("h1", class_="product_title")
    if not title_el:
        return None

    common_name_fr = title_el.get_text(strip=True)

    # The short description contains the scientific name and family.
    short_el = soup.select_one(
        ".woocommerce-product-details__short-description"
    )
    short_text = short_el.get_text("\n", strip=True) if short_el else ""

    # Try to extract scientific name and family from short description.
    scientific_name = None
    family = None
    for line in short_text.split("\n"):
        line = line.strip()
        if not line:
            continue
        if scientific_name is None:
            scientific_name = parse_scientific_name(line)
            family = parse_family(line)
            if scientific_name:
                break

    # The full description lives in the tab-description panel.
    tab_el = soup.find(id="tab-description")
    description = tab_el.get_text("\n", strip=True) if tab_el else ""

    return {
        "common name": common_name_fr,
        "scientific name": scientific_name,
        "family": family,
        "description": description,
    }


@define(frozen=True)
class SDPConverter(Converter):
    """Converter for La Société des Plantes fields."""

    locales: Locales = field(factory=partial(Locales.from_domain, "sdp"))

    def convert_item(self, key, value):
        dispatchers = {
            "common name": self.convert_ignore,
            "description": self.convert_string,
            "family": self.convert_string,
            "scientific name": self.convert_ignore,
        }
        return dispatchers.get(key, self.convert_string)(key, value)


@define(frozen=True)
class SDPModel:
    """La Société des Plantes model."""

    web: SDPWeb = field(factory=SDPWeb)
    converter: SDPConverter = field(factory=SDPConverter)

    def with_cache(self, storage):
        web = self.web.with_cache(storage)
        return evolve(self, web=web)

    def product_paths(self):
        """Return all French product paths from the sitemap."""
        xml_text = self.web.product_sitemap()
        return parse_sitemap_urls(xml_text)

    def get_plant(self, path):
        """Fetch and parse a single product page."""
        html = self.web.product_page(path)
        return parse_product(html)

    def all_plants(self):
        """Yield converted plant data for all products."""
        for path in self.product_paths():
            raw = self.get_plant(path)
            if raw is None or raw["scientific name"] is None:
                continue

            converted = self.converter.convert(raw)
            converted["scientific name"] = raw["scientific name"]

            common_name = raw["common name"]
            if common_name:
                converted[f"common name/{common_name}"] = True

            yield converted


@define(frozen=True)
class SDPIngestor:
    """Ingestor for La Société des Plantes."""

    model: SDPModel = field(factory=SDPModel)
    priority: Priority = field(factory=Priority)

    @classmethod
    def from_config(cls, config):
        """Instantiate SDPIngestor from config."""
        model = SDPModel().with_cache(config.storage)
        priority = LocationPriority("Quebec").with_cache(config.storage)
        return cls(model, priority)

    def fetch_all(self):
        count = 0
        for plant in self.model.all_plants():
            count += 1
            if count % 100 == 0:
                logger.info("SDP: ingested %d plants", count)
            yield DatabasePlant(plant, self.priority.weight)
        logger.info("SDP: ingested %d plants total", count)
